# -*- coding: utf-8 -*-
"""Partition function calculators and tabulators.

Notes
-----

Partition function calculators and tabulators

Calculators all derive from the same RovibPartitionFunction object,
and require a list of energies

Tabulators are more specific, and require a list of precalculated
partition functions at different temperature. PartFuncTIPS
uses the HITRAN hapi.py [1]_ module to interpolate Q for HITRAN species


Routine Listing
---------------

Partition functions:

- :class:`~radis.levels.partfunc.PartFuncTIPS`
- :class:`~radis.levels.partfunc.PartFunc_Dunham`

Which inherit from:

- :class:`~radis.levels.partfunc.RovibParFuncCalculator`
- :class:`~radis.levels.partfunc.RovibParFuncTabulator`

Which inherit from:

- :class:`~radis.levels.partfunc.RovibPartitionFunction`


See Also
--------

- :class:`~radis.levels.partfunc_cdsd.PartFuncCO2_CDSDcalc`
- :class:`~radis.levels.partfunc_cdsd.PartFuncCO2_CDSDtab`

References
----------

.. [1] `HAPI: The HITRAN Application Programming Interface <http://hitran.org/hapi>`_



-------------------------------------------------------------------------------
"""

# TODO: vectorize partition function calculations for different temperatures. Would need
# stuff like E = df.E.values.reshape((1,-1)), etc.

# TODO: store molecule_data.json in the H5 file metadata. If not done already.


import sys
from os.path import exists
from warnings import warn

import numpy as np
import pandas as pd
from numpy import exp

import radis
from radis.api.cache_files import load_h5_cache_file, save_to_hdf
from radis.api.kuruczapi import load_pf_Barklem2016
from radis.db.classes import (
    HITRAN_CLASS1,
    HITRAN_CLASS2,
    HITRAN_CLASS3,
    HITRAN_CLASS5,
    HITRAN_CLASS6,
    get_molecule,
    get_molecule_identifier,
)
from radis.lbl.labels import vib_lvl_name_hitran_class1, vib_lvl_name_hitran_class5
from radis.misc.basics import all_in
from radis.misc.debug import printdbg
from radis.misc.printer import printg
from radis.misc.progress_bar import ProgressBar
from radis.misc.warning import OutOfBoundError
from radis.phys.constants import hc_k  # ~ 1.44 cm.K


class RovibPartitionFunction(object):
    """General class from which all partition function calculators derive.

    Parameters
    ----------
    electronic_state: :class:`~radis.db.classes.ElectronicState`
        an :class:`~radis.db.classes.ElectronicState` object, which is
        defined in RADIS molecule database and contains spectroscopic data

    Notes
    -----
    Implementation:

    one partition function generator (RovibPartitionFunction) is generated
    per specie per isotope

    RovibPartitionFunction may differ in the way they build / fetch
    their list of states and the associated energies, but
    the .at(),  .at_noneq()  calls to calculate the partition
    function should be shared among all derived classes

    See Also
    --------
    :class:`~radis.levels.partfunc.RovibParFuncTabulator`,
    :class:`~radis.levels.partfunc.RovibParFuncCalculator`
    """

    def __init__(self):
        pass


# %% Subclasses :
# - Tabulator
# - Calculator


class RovibParFuncTabulator(RovibPartitionFunction):
    def __init__(self):
        super(RovibParFuncTabulator, self).__init__()

    def at(self, T, potential_lowering=None, **kwargs):
        r"""Get partition function at temperature T under equilibrium
        conditions, from tabulated data.

        Parameters
        ----------
        T: float
            equilibrium temperature
        potential_lowering: float
            The potential lowering in cm-1/Zeff**2, only relevant for Kurucz linelists which the partition function tables may have a dependence on this

        Returns
        -------
        Q: float
            partition function interpolated  at temperature T

        Examples
        --------
        See :ref:`online examples <label_examples_partition_functions>`

        See Also
        --------
        :py:class:`~radis.levels.partfunc.PartFuncTIPS`
        """

        # For compatibility with the syntax of RovibParFuncCalculator.at()
        update_populations = kwargs.pop("update_populations", False)
        if kwargs != {}:
            raise TypeError(
                "RovibParFuncTabulator.at() got an unexpected keyword: {0}".format(
                    kwargs.keys()
                )
            )
        if update_populations != False:
            raise ValueError(
                "Cannot use update_populations with tabulated partition functions. "
                + "Use a direct summation instead. See RovibParFuncCalculator"
            )

        # defined individually for each class Variants (one per database)
        if potential_lowering is not None:
            return self._at(T, potential_lowering)
        else:
            return self._at(T)

    def at_noneq(self, *args, **kwargs):
        raise ValueError(
            "Cannot calculate non equilibrium partition "
            + "functions from (equilibrium) tabulated values. Use a "
            + "Partition Function Calculator"
        )


class RovibParFuncCalculator(RovibPartitionFunction):
    """
    Parameters
    ----------
    electronic_state: :class:`~radis.db.classes.ElectronicState`
        an :class:`~radis.db.classes.ElectronicState` object, which is
        defined in RADIS molecule database and contains spectroscopic data.
        Energies are calculated with the :py:meth:`~radis.db.classes.ElectronicState.Erovib`
        method.

    Other Parameters
    ----------------
    mode: 'full summation', 'tabulation'
        calculation mode. ``'tabulation'`` is much faster but not all possible
        distributions are implemented. Default ``'full-summation'``

    See Also
    --------
    :py:class:`~radis.levels.partfunc.PartFunc_Dunham`,
    :py:meth:`~radis.db.classes.ElectronicState.Erovib`
    """

    def __init__(self, electronic_state, mode="full summation", verbose=False):

        super(RovibParFuncCalculator, self).__init__()

        ElecState = electronic_state

        try:
            ElecState.Erovib
        except AttributeError:
            raise AttributeError(
                "{0} has no energy function defined in RADIS".format(
                    ElecState.get_fullname()
                )
            )

        # Store
        self.ElecState = ElecState  # electronic state object
        self.molecule = ElecState.name  # molecule name
        self.isotope = ElecState.iso
        self.verbose = verbose

        # Line database
        # pandas Dataframe that holds all the lines
        self.df = pd.DataFrame({})
        # updated on inherited classes initialization

        if not mode in ["full summation", "tabulation"]:
            raise ValueError("Choose mode = one of 'full summation', 'tabulation'")
        if mode == "tabulation":
            try:
                import vaex

                vaex  # to avoid linting errors
            except ImportError:
                raise ImportError(
                    "On-the-fly tabulation of partition functions require Vaex. Install it. /!\ as of June 2024 Vaex is not available on Python>=3.11 "
                )

        self.mode = mode
        self._tab_at = None  # tabulated function
        self._tab_at_noneq = None  # tabulated function
        self._tab_at_noneq_3Tvib = None  # tabulated function
        self.N_bins = (
            200  # int: number of bins in tabulated mode. Change with `Z.N_bins = `
        )
        self.N_bins_scaling = lambda N_bins, Ndim: int(N_bins * (0.8 ** (Ndim - 1)))
        """ func int, int -> int
        Reduce number of Bins in each dimension ; in high dimensional spaces.
        This is justified by accuracy tests in :py:func:`radis.test.levels.test_partfunc.test_tabulated_partition_functions`
        """

    def at(self, T, update_populations=False):
        r"""Get partition function at temperature T under equilibrium
        conditions.

        Parameters
        ----------
        T: float
            equilibrium temperature

        Other Parameters
        ----------------
        update_populations: boolean
            if ``True``, store calculated populations in energy level list
            Default ``False``

        Returns
        -------
        Q: float
            partition function calculated at temperature T

        Examples
        --------
        See :ref:`online examples <label_examples_partition_functions>`

        See Also
        --------
        :meth:`~radis.levels.partfunc.RovibPartitionFunction.at_noneq`,
        :meth:`~radis.levels.partfunc.RovibPartitionFunction.at_noneq_3Tvib`
        """
        if __debug__:
            printdbg(
                "called RovibPartitionFunction.at(T={0}K, ".format(T)
                + "update_populations={0})".format(update_populations)
                + f". mode = {self.mode}"
            )

        # Check inputs
        assert isinstance(update_populations, bool)

        if self.mode == "full summation":
            return self._eq_full_summation(T=T, update_populations=update_populations)
        elif self.mode == "tabulation":
            if update_populations:
                raise ValueError(
                    "Cannot update populations of individual levels with `tabulation` mode. Choose `update_populations=False` or `mode='full summation'`"
                )
            return self._eq_tabulation_eval(T=T)

    def _eq_tabulation_setup(self, N_bins):
        r"""Bins all levels into an ``E`` grid

        Parameters
        ----------
        N_bins: int
            reset it by editing the class attribute `Z.N_bins = `

        See Also
        --------
        :py:func:`~radis.levels.partfunc._noneq_tabulation_eval`
        """
        # Note @dev : this is a Vaex only feature
        shape = N_bins
        if self.verbose >= 3:
            print(f"Tabulation eq partition functions with : shape = {shape}")

        # Get variables
        import vaex  # import delayed until now (takes ~2s to import)

        df = vaex.from_pandas(self.df)
        # Vaex process may get stuck in Spyder IDE. https://github.com/radis/radis/issues/338
        # Temp fix : >>> df.executor.async_method = "awaitio"     (doesn't always work here)

        epsilon = 1e-4  # prevent log(0)
        df["logE"] = np.log(df["E"] + epsilon)  # to bin on a log grid
        df["gtot"] = (
            df["grot"] * df["gvib"]
        )  # note that this column is "lazy" and only evaluated at runtime
        E_bins = df.mean("E", binby="logE", shape=shape)
        g_bins = df.sum("gtot", binby="logE", shape=shape)

        # drop empty
        E_bins = E_bins[g_bins > 0]
        g_bins = g_bins[g_bins > 0]

        self._tab_at = lambda T: (g_bins * exp(-hc_k * E_bins / T)).sum(axis=0)
        # Also save parameters to trigger a re-tabulation if they change:
        self._tab_N_bins = N_bins

    def _eq_tabulation_eval(self, T):
        r"""Computes partition function using tabulated grid

        See Also
        --------
        :py:func:`~radis.levels.partfunc._noneq_tabulation_setup`"""

        N_bins = self.N_bins  # reset it by editing the class attribute `Z.N_bins = `
        N_bins = self.N_bins_scaling(N_bins, 1)

        if self._tab_at is None or N_bins != self._tab_N_bins:
            # Tabulate or re-tabulate:
            self._eq_tabulation_setup(N_bins=N_bins)

        return self._tab_at(T)

    def _eq_full_summation(self, T, update_populations=False):

        # Initialize
        df = self.df
        if "g" in df.columns:
            g = df.g
        elif all_in(["gvib", "grot"], df.columns):
            g = df.gvib * df.grot
        else:
            raise ValueError(
                "either g, or gvib+grot must be defined to "
                + "calculate total degeneracy. Got: {0}".format(list(df.keys()))
            )

        # Calculate

        nQ = g * exp(-hc_k * df.E / T)
        Q = nQ.sum()

        # Update energy level table with populations (doesnt
        # cost much and can be used to plot populations afterwards)
        # ... 'n'
        if update_populations:
            df["n"] = nQ / Q

        return Q

    def at_noneq(
        self,
        Tvib,
        Trot,
        overpopulation=None,
        vib_distribution="boltzmann",
        rot_distribution="boltzmann",
        returnQvibQrot=False,
        update_populations=False,
    ):
        r"""Calculate Partition Function under non equilibrium (Tvib, Trot),
        with boltzmann/treanor distributions and overpopulations as specified
        by the user.

        Parameters
        ----------
        Tvib, Trot: float
            vibrational & rotational temperatures (K)
        overpopulation: dict, or ``None``
            dict of overpopulated levels: ``{'level':over_factor}``
        vib_distribution: ``'boltzmann'``, ``'treanor'``
            distribution of vibrational levels
        rot_distribution: ``'boltzmann'``
            distribution of rotational levels
        returnQvibQrot: boolean
            cf output

        Other Parameters
        ----------------
        update_populations: boolean
            if ``True``, store calculated populations in energy level list.
            Default ``False``

        Returns
        -------
        Q: float
            partition function calculated at non eq temperatures
        Q, Qvib, dfQrot: float, float, pandas table (if returnQvibQrot)
            total partition function, vibrational partition function,
            and table of rotational partition functions for each vibrational
            state (note that all Qrot are not necessarily the same
            for all vibrational levels)

        Examples
        --------
        See :ref:`online examples <label_examples_partition_functions>`

        See Also
        --------
        :meth:`~radis.levels.partfunc.RovibPartitionFunction.at`,
        :meth:`~radis.levels.partfunc.RovibPartitionFunction.at_noneq_3Tvib`
        """
        if __debug__:
            printdbg(
                "called RovibPartitionFunction.atnoneq"
                + "(Tvib={0}K, Trot={1}K, ... )".format(Tvib, Trot)
                + "update_populations={0})".format(update_populations)
                + f". mode = {self.mode}"
            )

        # Check inputs, initialize
        if overpopulation is None:
            overpopulation = {}
        if overpopulation != {}:
            if not returnQvibQrot:
                raise ValueError(
                    "When using overpopulation, partition function "
                    + "must be calculated with returnQvibQrot=True. Set ``PartitionFunctionCalculator.at_noneq(..., returnQvibQrot=True)``, or ``SpectrumFactory.misc.export_rovib_fraction = True``"
                )
        assert vib_distribution in ["boltzmann", "treanor"]
        assert rot_distribution in ["boltzmann"]
        if vib_distribution == "boltzmann":
            if not "Evib" in list(self.df.keys()):
                raise ValueError(
                    "Evib must be defined to calculate non-equilibrium "
                    + "partition functions"
                )
        elif vib_distribution == "treanor":
            if not all_in(["Evib_h", "Evib_a"], list(self.df.keys())):
                raise ValueError(
                    "Evib_h and Evib_a must be defined to calculate non-equilibrium "
                    + "partition functions with treanor distribution"
                )
        if rot_distribution == "boltzmann":
            if not "Erot" in list(self.df.keys()):
                raise ValueError(
                    "Evib and Erot must be defined to calculate non-equilibrium "
                    + "partition functions"
                )

        if self.mode == "full summation":
            return self._noneq_full_summation(
                Tvib=Tvib,
                Trot=Trot,
                overpopulation=overpopulation,
                vib_distribution=vib_distribution,
                rot_distribution=rot_distribution,
                returnQvibQrot=returnQvibQrot,
                update_populations=update_populations,
            )
        elif self.mode == "tabulation":
            if update_populations:
                raise ValueError(
                    "Cannot update populations of individual levels with `tabulation` mode. Choose `update_populations=False` or `mode='full summation'`"
                )
            return self._noneq_tabulation_eval(
                Tvib=Tvib,
                Trot=Trot,
                overpopulation=overpopulation,
                vib_distribution=vib_distribution,
                rot_distribution=rot_distribution,
                returnQvibQrot=returnQvibQrot,
            )

    def _noneq_tabulation_setup(self, N_bins, vib_distribution, rot_distribution):
        r"""Bins all levels into an Evib and Erot grid

        Parameters
        ----------
        N_bins: int
            reset it by editing the class attribute `Z.N_bins = `

        See Also
        --------
        :py:func:`~radis.levels.partfunc._noneq_tabulation_eval`
        """
        shape = N_bins, N_bins
        if self.verbose >= 3:
            print(f"Tabulation noneq partition functions with : shape = {shape}")

        # Get variables
        import vaex  # import delayed until now (takes ~2s to import)

        df = vaex.from_pandas(self.df)
        # Vaex process may get stuck in Spyder IDE. https://github.com/radis/radis/issues/338
        # Temp fix : >>> df.executor.async_method = "awaitio"     (doesn't always work here)

        epsilon = 1e-4  # prevent log(0)
        df["logEvib"] = np.log(df["Evib"] + epsilon)  # to bin on a log grid
        df["logErot"] = np.log(df["Erot"] + epsilon)  # to bin on a log grid
        df["gtot"] = df["grot"] * (
            df["gvib"]
        )  # note that this column is "lazy" and only evaluated at runtime
        # Evib_bins_neq = df.mean(
        #     "Evib", binby=["logEvib", "logErot"], shape=(N_bins, N_bins)
        # )
        # Erot_bins_neq = df.mean(
        #     "Erot", binby=["logEvib", "logErot"], shape=(N_bins, N_bins)
        # )
        Evib_bins_neq, Erot_bins_neq = df.mean(
            ["Evib", "Erot"], binby=["logEvib", "logErot"], shape=shape
        )
        g_bins_neq = df.sum("gtot", binby=["logEvib", "logErot"], shape=shape)
        # drop empty
        Evib_bins_neq = Evib_bins_neq[g_bins_neq > 0]
        Erot_bins_neq = Erot_bins_neq[g_bins_neq > 0]
        g_bins_neq = g_bins_neq[g_bins_neq > 0]

        self._tab_at_noneq = lambda Tvib, Trot: (
            g_bins_neq
            * exp(-hc_k * Evib_bins_neq / Tvib)
            * exp(-hc_k * Erot_bins_neq / Trot)
        ).sum(axis=0)
        # Also save parameters to trigger a re-tabulation if they change:
        self._tab_N_bins = N_bins
        self._tab_vib_distribution = vib_distribution
        self._tab_rot_distribution = rot_distribution

    def _noneq_tabulation_eval(
        self,
        Tvib,
        Trot,
        overpopulation=None,
        vib_distribution="boltzmann",
        rot_distribution="boltzmann",
        returnQvibQrot=False,
    ):
        r"""Computes partition function using tabulated grid

        See Also
        --------
        :py:func:`~radis.levels.partfunc._noneq_tabulation_setup`"""

        if len(overpopulation) > 0:
            raise NotImplementedError
        if vib_distribution != "boltzmann":
            raise NotImplementedError
        if rot_distribution != "boltzmann":
            raise NotImplementedError
        if returnQvibQrot != False:
            raise NotImplementedError

        N_bins = self.N_bins  # reset it by editing the class attribute `Z.N_bins = `
        N_bins = self.N_bins_scaling(N_bins, 2)

        if (
            self._tab_at_noneq is None
            or N_bins != self._tab_N_bins
            or self._tab_vib_distribution != vib_distribution
            or self._tab_rot_distribution != rot_distribution
        ):
            # Tabulate or re-tabulate:
            self._noneq_tabulation_setup(
                N_bins=N_bins,
                vib_distribution=vib_distribution,
                rot_distribution=rot_distribution,
            )

        return self._tab_at_noneq(Tvib, Trot)

    def _noneq_full_summation(
        self,
        Tvib,
        Trot,
        overpopulation=None,
        vib_distribution="boltzmann",
        rot_distribution="boltzmann",
        returnQvibQrot=False,
        update_populations=False,
    ):
        r"""Computes partition function by summing over all levels"""

        # Get variables
        df = self.df
        gvib = df.gvib  # self.gvib(M, I)
        grot = df.grot
        # Calculate

        # ... mode: Trot, Tvib, + overpopulation
        if returnQvibQrot:

            if not "viblvl" in self.df:
                raise KeyError(
                    "To return Qrot and Qvib vibrational levels must be "
                    + "identified in the database. Lookup function "
                    + "add_bands in radis.lbl.bands"
                )

            # ... Vibrational populations
            if vib_distribution == "boltzmann":
                df["nvibQvib"] = gvib * exp(-df.Evib * hc_k / Tvib)
            elif vib_distribution == "treanor":
                df["nvibQvib"] = gvib * exp(
                    -hc_k * (df.Evib_h / Tvib + df.Evib_a / Trot)
                )
            else:
                raise NotImplementedError
            # ... Rotational populations
            if rot_distribution == "boltzmann":
                df["nrotQrot"] = grot * exp(-df.Erot * hc_k / Trot)
            else:
                raise NotImplementedError

            if overpopulation != {}:
                # If overpopulation, first check all levels exist
                levels = df.viblvl.unique()
                for viblvl in overpopulation.keys():
                    if not viblvl in levels:
                        raise ValueError(
                            "Level {0} not in energy levels database".format(viblvl)
                        )
                        # could be a simple warning too
                # Add overpopulations (so they are taken into account in the partition function)
                for viblvl, ov in overpopulation.items():
                    if ov != 1:
                        df.loc[df.viblvl == viblvl, "nvibQvib"] *= ov
                        # TODO: add warning if empty? I dont know how to do it without
                        # an extra lookup though.

            # Calculate sum of levels
            nQ = df.nvibQvib * df.nrotQrot
            Q = nQ.sum()

            # Group by vibrational level, and get level-dependant
            # quantities such as vib degeneracy, Evib, etc.
            dgb = df.groupby(["viblvl"], as_index=True)
            # perf: 345 ms ± 5.88 ms on test case
            viblvl_Qrot = dgb["nrotQrot"].sum()
            # (all the same in one vib group)
            viblvl_nvibQvib = dgb["nvibQvib"].first()
            Qvib = viblvl_nvibQvib.sum()  # (here we only sum once per vib group)
            # (all the same in one vib group)
            viblvl_Evib = dgb["Evib"].first()
            # (all the same in one vib group)
            viblvl_gvib = dgb["gvib"].first()

            # Energies, degeneracies, populations for each vibrational band
            dfQrot = pd.DataFrame({"Qrot": viblvl_Qrot})
            dfQrot["nvib"] = viblvl_nvibQvib / Qvib
            dfQrot["Evib"] = viblvl_Evib
            dfQrot["gvib"] = viblvl_gvib

            # Update energy level table with populations (doesnt
            # cost much and can be used to plot populations afterwards)
            # Adds: 'nvib', 'n', 'nrot', 'Qrot'
            if update_populations:
                df["nvib"] = df.nvibQvib / Qvib
                df["n"] = nQ / Q

                # get rotational populations
                # ... reindexing dfQrot to get a direct access by viblvl
                dfQrot_dict = dict(list(zip(dfQrot.index, dfQrot.Qrot)))

                # ... Add Qrot
                df_viblvl = df.set_index(["viblvl"], inplace=False)
                df["Qrot"] = df_viblvl.index.map(dfQrot_dict.get).values

                # ... infer nrot
                df["nrot"] = df.nrotQrot / df.Qrot

            # Check that partition functions are valid
            if __debug__:  # discarded if running with python -O
                # the calculation below is valid without Born-Oppenheimer, however
                # it assumes Boltzmann distribution + no overpopulation
                #                Qrovib = ((dfQrot.gvib*exp(-dfQrot.Evib*hc_k/Tvib))*dfQrot.Qrot).sum()
                # general case:
                Qrovib = (dfQrot.nvib * Qvib * dfQrot.Qrot).sum()
                if not np.isclose(Q, Qrovib):
                    raise ValueError(
                        "Rovibrational partition function ({0:.2f}) doesnt ".format(Q)
                        + "match value recomputed from vibrational and rotational "
                        + "partition functions ({0:.2f}). Check how Evib and Erot ".format(
                            Qrovib
                        )
                        + "are defined in your Energy Database"
                    )

            # Clean
            del df["nrotQrot"]
            del df["nvibQvib"]

            return Q, Qvib, dfQrot

        # ... mode: Trot, Tvib, no overpopulation, no vib band details
        else:  # slightly faster, but doesnt return nvib nor Qvib

            g = gvib * grot

            if vib_distribution == "boltzmann" and rot_distribution == "boltzmann":
                nQ = g * exp(-hc_k * (df.Evib / Tvib + df.Erot / Trot))
            elif vib_distribution == "treanor" and rot_distribution == "boltzmann":
                nQ = g * exp(-hc_k * (df.Evib_h / Tvib + (df.Evib_a + df.Erot) / Trot))
            else:
                raise NotImplementedError

            Q = nQ.sum()

            # Update energy level table with populations (doesnt
            # cost much and can be used to plot populations afterwards)
            # ... add: 'n'
            if update_populations:
                df["n"] = nQ / Q

            return Q

    def at_noneq_3Tvib(
        self,
        Tvib,
        Trot,
        overpopulation=None,
        vib_distribution="boltzmann",
        rot_distribution="boltzmann",
        returnQvibQrot=False,
        update_populations=False,
    ):
        r"""Calculate Partition Function under non equilibrium ((Tvib1, Tvib2,
        Tvib3), Trot), with boltzmann/treanor distributions and overpopulations
        as specified by the user.

        Dedicated function for 3 Tvib mode

        Parameters
        ----------

        Tvib, Trot: (float, float, float), float
            vibrational & rotational temperatures (K)
        overpopulation: dict, or ``None``
            dict of overpopulated levels: ``{'level':over_factor}``
        vib_distribution: ``'boltzmann'``, ``'treanor'``
        rot_distribution: ``'boltzmann'``
        returnQvibQrot: boolean
            cf output
        update_populations: boolean
            if ``True``, saves populations for calculated temperature in PartitionFunction
            dataframe. Default ``False``

        Returns
        -------
        Q: float
            partition function calculated at non eq temperatures
        Q, Qvib, dfQrot: float, float, pandas table (if returnQvibQrot)
            total partition function, vibrational partition function,
            and table of rotational partition functions for each vibrational
            state (note that all Qrot are not necessarily the same
            for all vibrational levels)

        See Also
        --------
        :meth:`~radis.levels.partfunc.RovibPartitionFunction.at`,
        :meth:`~radis.levels.partfunc.RovibPartitionFunction.at_noneq`
        """
        if __debug__:
            printdbg(
                "called RovibPartitionFunction.at_noneq_3Tvib"
                + "(Tvib={0}K, Trot={1}K)".format(Tvib, Trot)
                + "update_populations={0})".format(update_populations)
                + f". mode = {self.mode}"
            )
        #                               'overpopulation={0}, vib_distribution={1}'.format(overpopulation, vib_distribution)+\
        #                               'rot_distribution={0})'.format(rot_distribution)

        # Check inputs
        if overpopulation is None:
            overpopulation = {}
        #        else:
        #            if not returnQvibQrot:
        #                raise ValueError('When using overpopulation partition function '+\
        #                                 'must be calculated with returnQvibQrot')
        assert vib_distribution in ["boltzmann", "treanor"]
        assert rot_distribution in ["boltzmann"]
        if vib_distribution == "boltzmann":
            if not all_in(["Evib1", "Evib2", "Evib3"], list(self.df.keys())):
                raise ValueError(
                    "Evib1, Evib2, Evib3 must be defined to calculate non-equilibrium "
                    + "partition functions with Tvib1, Tvib2, Tvib3. Got {0}".format(
                        list(self.df.keys())
                    )
                )
        elif vib_distribution == "treanor":
            if not all_in(
                ["Evib1_h", "Evib1_a", "Evib2_h", "Evib2_a", "Evib3_h", "Evib3_a"],
                list(self.df.keys()),
            ):
                raise ValueError(
                    "Evib1_h, Evib1_a, Evib2_h, Evib2_a, Evib3_h, Evib3_a "
                    + "must be defined to calculate non-equilibrium "
                    + "partition functions with treanor distribution "
                    + "and Tvib1, Tvib2, Tvib3"
                )
        if rot_distribution == "boltzmann":
            if not "Erot" in list(self.df.keys()):
                raise ValueError(
                    "Evib and Erot must be defined to calculate non-equilibrium "
                    + "partition functions"
                )

        # Calculate
        if self.mode == "full summation":
            return self._noneq_3Tvib_full_summation(
                Tvib=Tvib,
                Trot=Trot,
                overpopulation=overpopulation,
                vib_distribution=vib_distribution,
                rot_distribution=rot_distribution,
                returnQvibQrot=returnQvibQrot,
                update_populations=update_populations,
            )
        elif self.mode == "tabulation":
            if update_populations:
                raise ValueError(
                    "Cannot update populations of individual levels with `tabulation` mode. Choose `update_populations=False` or `mode='full summation'`"
                )
            return self._noneq_3Tvib_tabulation_eval(
                Tvib=Tvib,
                Trot=Trot,
                overpopulation=overpopulation,
                vib_distribution=vib_distribution,
                rot_distribution=rot_distribution,
                returnQvibQrot=returnQvibQrot,
            )

    def _noneq_3Tvib_tabulation_setup(self, N_bins, vib_distribution, rot_distribution):
        r"""Bins all levels into an Evib and Erot grid

        Parameters
        ----------
        N_bins: int
            reset it by editing the class attribute `Z.N_bins = `

        See Also
        --------
        :py:func:`~radis.levels.partfunc._noneq_tabulation_eval`
        """
        shape = N_bins, N_bins, N_bins
        if self.verbose >= 3:
            print(f"Tabulation noneq 3Tvib partition functions with : shape = {shape}")

        # Get variables
        import vaex  # import delayed until now (takes ~2s to import)

        df = vaex.from_pandas(self.df)
        # Vaex process may get stuck in Spyder IDE. https://github.com/radis/radis/issues/338
        # Temp fix : >>> df.executor.async_method = "awaitio"     (doesn't always work here)

        epsilon = 1e-4  # prevent log(0)

        if vib_distribution == "boltzmann" and rot_distribution == "boltzmann":

            if "Evib12" not in df:
                df["Evib12"] = df["Evib1"] + df["Evib2"]
            df["logEvib12"] = np.log(df["Evib12"] + epsilon)  # to bin on a log grid
            df["logEvib3"] = np.log(df["Evib3"] + epsilon)  # to bin on a log grid
            df["logErot"] = np.log(df["Erot"] + epsilon)  # to bin on a log grid
            df["gtot"] = df["grot"] * (
                df["gvib"]
            )  # note that this column is "lazy" and only evaluated at runtime

            Evib12_bins_neq, Evib3_bins_neq, Erot_bins_neq = df.mean(
                ["Evib12", "Evib3", "Erot"],
                binby=["logEvib12", "logEvib3", "logErot"],
                shape=shape,
            )
            g_bins_neq = df.sum(
                "gtot",
                binby=["logEvib12", "logEvib3", "logErot"],
                shape=shape,
            )

            # drop empty
            Evib12_bins_neq = Evib12_bins_neq[g_bins_neq > 0]
            Evib3_bins_neq = Evib3_bins_neq[g_bins_neq > 0]
            Erot_bins_neq = Erot_bins_neq[g_bins_neq > 0]
            g_bins_neq = g_bins_neq[g_bins_neq > 0]

            self._tab_at_noneq_3Tvib = lambda Tvib, Trot: (
                g_bins_neq
                * exp(-hc_k * Evib12_bins_neq / Tvib[0])
                * exp(-hc_k * Evib3_bins_neq / Tvib[2])
                * exp(-hc_k * Erot_bins_neq / Trot)
            ).sum(axis=0)

        elif vib_distribution == "treanor" and rot_distribution == "boltzmann":

            if "Evib12_h" not in df:
                df["Evib12_h"] = df["Evib1_h"] + df["Evib2_h"]
            if "E_anharmonic" not in df:
                df["E_anharmonic"] = df["Evib1_a"] + df["Evib2_a"] + df["Erot"]
            df["logEvib12_h"] = np.log(df["Evib12_h"] + epsilon)  # to bin on a log grid
            df["logEvib3_h"] = np.log(df["Evib3_h"] + epsilon)  # to bin on a log grid
            df["E_anharmonic"] = np.log(
                df["E_anharmonic"] + epsilon
            )  # to bin on a log grid
            df["gtot"] = df["grot"] * df["gvib"]

            Evib12_h_bins_neq, Evib3_h_bins_neq, E_anharmonic_bins_neq = df.mean(
                ["Evib12_h", "Evib3_h", "E_anharmonic"],
                binby=["Evib12_h", "Evib3_h", "E_anharmonic"],
                shape=shape,
            )
            g_bins_neq = df.sum(
                "gtot",
                binby=["Evib12_h", "Evib3_h", "E_anharmonic"],
                shape=shape,
            )

            # drop empty
            Evib12_h_bins_neq = Evib12_h_bins_neq[g_bins_neq > 0]
            Evib3_h_bins_neq = Evib3_h_bins_neq[g_bins_neq > 0]
            E_anharmonic_bins_neq = E_anharmonic_bins_neq[g_bins_neq > 0]
            g_bins_neq = g_bins_neq[g_bins_neq > 0]

            self._tab_at_noneq_3Tvib = lambda Tvib, Trot: (
                g_bins_neq
                * exp(-hc_k * Evib12_h_bins_neq / Tvib[0])
                * exp(-hc_k * Evib3_h_bins_neq / Tvib[2])
                * exp(-hc_k * E_anharmonic_bins_neq / Trot)
            ).sum(axis=0)

        else:
            raise NotImplementedError(
                f"vib_distribution: {vib_distribution}, rot_distribution: {rot_distribution}"
            )

        # Also save parameters to trigger a re-tabulation if they change:
        self._tab_N_bins = N_bins
        self._tab_vib_distribution = vib_distribution
        self._tab_rot_distribution = rot_distribution

    def _noneq_3Tvib_tabulation_eval(
        self,
        Tvib,
        Trot,
        overpopulation=None,
        vib_distribution="boltzmann",
        rot_distribution="boltzmann",
        returnQvibQrot=False,
    ):
        r"""Computes partition function using tabulated grid

        See Also
        --------
        :py:func:`~radis.levels.partfunc._noneq_tabulation_setup`"""

        Tvib1, Tvib2, Tvib3 = Tvib
        if Tvib1 != Tvib2:
            raise NotImplementedError(
                "Tabulated mode only implemented for Tvib1 = Tvib2. Use mode='full summation'"
            )
        if len(overpopulation) > 0:
            raise NotImplementedError
        if rot_distribution != "boltzmann":
            raise NotImplementedError
        if returnQvibQrot != False:
            raise NotImplementedError

        N_bins = self.N_bins  # reset it by editing the class attribute `Z.N_bins = `
        N_bins = self.N_bins_scaling(N_bins, 3)  # 3 temperatures only (T12, T3, Trot)

        if (
            self._tab_at_noneq_3Tvib is None
            or N_bins != self._tab_N_bins
            or self._tab_vib_distribution != vib_distribution
            or self._tab_rot_distribution != rot_distribution
        ):
            # Tabulate or re-tabulate:
            self._noneq_3Tvib_tabulation_setup(
                N_bins=N_bins,
                vib_distribution=vib_distribution,
                rot_distribution=rot_distribution,
            )

        return self._tab_at_noneq_3Tvib(Tvib, Trot)

    def _noneq_3Tvib_full_summation(
        self,
        Tvib,
        Trot,
        overpopulation=None,
        vib_distribution="boltzmann",
        rot_distribution="boltzmann",
        returnQvibQrot=False,
        update_populations=False,
    ):
        r"""Computes partition function by summing over all levels"""

        # Get variables
        Tvib1, Tvib2, Tvib3 = Tvib
        df = self.df
        gvib = df.gvib
        grot = df.grot

        # Calculate
        if overpopulation:
            raise NotImplementedError("overpopulation not implemented for 3 Tvib mode")

        # ... mode: Trot, Tvib, + overpopulation
        if returnQvibQrot:
            raise NotImplementedError("returnQvibQrot not implemented for 3 Tvib mode")

        # ... mode: Trot, Tvib, no overpopulation, no vib band details
        else:  # much faster, but doesnt return nvib nor Qvib

            g = gvib * grot

            if vib_distribution == "boltzmann" and rot_distribution == "boltzmann":
                nQ = g * exp(
                    -hc_k
                    * (
                        df.Evib1 / Tvib1
                        + df.Evib2 / Tvib2
                        + df.Evib3 / Tvib3
                        + df.Erot / Trot
                    )
                )
            elif vib_distribution == "treanor" and rot_distribution == "boltzmann":
                nQ = g * exp(
                    -hc_k
                    * (
                        df.Evib1_h / Tvib1
                        + df.Evib2_h / Tvib2
                        + df.Evib3_h / Tvib3
                        + (df.Evib1_a + df.Evib2_a + df.Evib3_a + df.Erot) / Trot
                    )
                )
            else:
                raise NotImplementedError

            Q = nQ.sum()

            # Update energy level table with populations
            # ... add: 'n'
            if update_populations:
                df["n"] = nQ / Q

            return Q

    def reset_populations(self):
        r"""Discard computed populations of all energy levels.

        To call on every RovibrationalPartitionFunction object before
        each new spectrum calculation
        """

        for k in ["nvib", "n", "nrot"]:
            if k in self.df:
                del self.df[k]

    # %% Methods to get populations of all states

    def _get_vib_populations(self):
        r"""Return vibrational populations for all levels featured in given line
        set."""

        df = self.df

        return df.drop_duplicates("viblvl")

    def _get_rovib_populations(self):
        r"""Return rovibrational populations for all levels featured in the
        energy levels list df.

        Notes
        -----

        assumes a complete rovibrational assignment but no hyperfine assignment
        (i.e: all energy levels are returned!). If hyperfine assignment is given,
        this method should be modified to return only the ``roviblvl`` unique
        keys
        """

        df = self.df

        return df


# %% Variants of Tabulated partition functions (interpolate)


class PartFuncExoMol(RovibParFuncTabulator):
    """Return partition function using interpolation of tabulated values.

    Parameters
    ----------
    name: str
        exomol isotope full name
    T_range, Q_range: array
        initial references values to be tabulated.

    Examples
    --------

    One can use :py:func:`~radis.io.exomol.fetch_exomol` which has the option to
    return a :py:class:`~radis.levels.partfunc.PartFuncExoMol` object ::

        from radis.io.exomol import fetch_exomol
        from radis.api.exomolapi import get_exomol_database_list

        print(get_exomol_database_list("SiO", "28Si-16O"))  # 'EBJT" is one database
        _, _, Z_exomol = fetch_exomol("SiO", "EBJT", return_local_path=True, return_partition_function=True)


        # Get temperature at 2000 K
        print(Z_exomol.at(2000))


    See Also
    --------
    :py:func:`~radis.io.exomol.fetch_exomol`
    """

    def __init__(self, name, T_range, Q_range):

        super(PartFuncExoMol, self).__init__()

        self.name = name
        self.T_range = T_range
        self.Q_range = Q_range

    def _at(self, T):
        return np.interp(T, self.T_range, self.Q_range)


class PartFuncKurucz(RovibParFuncTabulator):
    """Return partition function using interpolation of tabulated values of local file kuruczpartfn.txt"""

    def __init__(self, species, pfpath=None):
        super(PartFuncKurucz, self).__init__()
        # Load data in constructor
        self.species = species
        self.pfTdat, self.pfdat = load_pf_Barklem2016()
        # Locate the row for the specific atom and ionization state
        try:
            pf_atom = self.pfdat.loc[f"{species}"]
        except KeyError:
            warn(
                "Note: the partition functions from Barklem & Collet (2016) don't include this species."
            )
            self.pf_values = self.pfT_values = None
        else:
            self.pfT_values = self.pfTdat.values.flatten().astype(
                float
            )  # namespace["pfT_values"]
            self.pf_values = pf_atom.values.astype(float)

        if pfpath:
            self.partfn = pd.read_hdf(pfpath)
        else:
            self.partfn = None

    def _at(self, T, potential_lowering=None):
        # Interpolate to find the partition function at the desired temperature
        if potential_lowering is not None and self.partfn is not None:
            Temp = self.partfn["T"]
            try:
                Qvals = self.partfn[f"{potential_lowering}"]
            except KeyError:
                print("Available values of potential lowering:")
                print(
                    *self.partfn.columns.difference(["T"], sort=False).to_list(),
                    sep=", ",
                )
                raise
            addmsg = "the Kurucz partition function. You can try the Barklem & Collet (2016) partition function by setting `potential_lowering=None`."
        else:
            if self.partfn is None:
                if potential_lowering is not None:
                    warn(
                        "Table of partition functions by potential lowering not available for this species - using Barklem & Collet (2016) instead with just the temperature"
                    )
                addmsg = "Barklem & Collet (2016)."
                addmsg2 = ", nor are any dedicated tables available for this species."
            else:
                addmsg = "Barklem & Collet (2016). You can try the Kurucz partition function by setting a potential lowering (e.g. `potential_lowering=-500`)."
                addmsg2 = ". Please specify the potential lowering."
            if self.pf_values is None:
                raise Exception(
                    "The partition functions from Barklem & Collet (2016) don't include this species"
                    + addmsg2
                )

            Temp = self.pfT_values
            Qvals = self.pf_values

        if T < Temp.min() or T > Temp.max():
            raise ValueError(
                f"The temperature {T} K is outside the tabulated range of the partition functions [{Temp.min()}, {Temp.max()}] K for "
                + addmsg
            )
        return np.interp(T, Temp, Qvals)


class PartFuncTIPS(RovibParFuncTabulator):
    """Return partition function using interpolation of tabulated values
    using the TIPS program [TIPS-2020]_

    Parameters
    ----------
    M: int
        molecule id
    I: int
        isotope identifier
    path: str
        path to ``hapi.py``. If None, the hapi package from PyPI is used.

    Examples
    --------
    ::

        from radis.levels.partfunc import PartFuncTIPS
        from radis.db.classes import get_molecule_identifier

        M = get_molecule_identifier('N2O')
        iso=1

        Q = PartFuncTIPS(M, iso)
        print(Q.at(T=1500))

    See :ref:`online examples <label_examples_partition_functions>` for more.

    .. minigallery:: radis.levels.partfunc.PartFuncTIPS

    References
    ----------
    Partition function are retrieved from [TIPS-2020]_ through [HAPI]_  (Hitran Python Interface) using
    partitionSum(M,I,T)
    """

    def __init__(self, M, I, path=None, verbose=True):

        super(PartFuncTIPS, self).__init__()

        self.verbose = verbose

        if path is not None:
            partitionSum = self.import_from_file(path)
        else:
            # Use TIPS-2017 through HAPI
            if not verbose:
                # import HAPI without printing details if verbose=False.
                from contextlib import redirect_stdout

                with redirect_stdout(None):
                    from hapi import HAPI_VERSION, partitionSum
            else:
                from hapi import HAPI_VERSION, partitionSum

                if self.verbose >= 2:
                    printg("HAPI version: %s" % HAPI_VERSION)

        # Check inputs
        if isinstance(M, str):
            M = get_molecule_identifier(M)
        if type(M) is not int:
            raise TypeError("Molecule id must be int: got {0} ({1})".format(M, type(M)))
        if type(I) not in [int, np.int64]:
            raise TypeError(
                "Isotope number must be int: got {0} ({1})".format(I, type(I))
            )

        self.partitionSum = partitionSum
        self.M = M
        self.molecule = get_molecule(M)
        self.I = I

    def import_from_file(self, path):
        r"""Import hapi.py from a given file (in case user wants to specify a
        different HAPI version than the one embedded in RADIS)"""
        if sys.version == 2:
            import imp

            hapi = imp.load_source("hapi", path)
        else:  # Python 3
            import importlib.util

            spec = importlib.util.spec_from_file_location("hapi", path)
            hapi = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(hapi)
        return hapi.partitionSum

    def _at(self, T):
        r"""Get partition function of species M, isotope I at temperature T.

        Called by :meth:`radis.levels.partfunc.RovibParFuncTabulator.at`
        """
        try:
            return self.partitionSum(self.M, self.I, T)
        except Exception as err:
            if "TIPS" in str(err):  # improve error message
                from radis.db import MOLECULES_LIST_NONEQUILIBRIUM

                if self.molecule in MOLECULES_LIST_NONEQUILIBRIUM:
                    tip = "\n💡 You may circumvent this by computing the partition function yourself using RADIS built-in spectroscopic constants. Check their range of validity, though ! If you decide to do so, use a nonequilibrium calculation with `Tvib=Trot`"
                else:
                    tip = ""
                raise OutOfBoundError(
                    "Partition functions not computed in TIPS "
                    + f"for molecule {self.molecule}, isotope {self.I}. "
                    + f"Max range : {str(err)}"
                    + tip
                )
            elif isinstance(err, KeyError):
                molecule = self.molecule
                # In current version, GEISA parser uses HAPI partition function, which does not
                # support some of GEISA's isotopes, as listed in GEISA_nw_iso below:
                GEISA_ns_iso = {
                    "CO2": [0],
                    "H2O": [8, 9],
                    "N2O": [0],
                    "NO2": [2],
                }
                raise KeyError(
                    "KeyError spotted! "
                    + f"If you are computing GEISA spectra, this result might be because of an "
                    + f"unsupported isotope. Currently isotope ID {GEISA_ns_iso[molecule]} "
                    + f"of molecule {molecule} is not supported by HAPI partitional function, "
                    + "thus stopping this spectrum calculation. Please select other isotopes "
                    + "lists such as isotope='1,2,3'."
                ) from err

            else:
                raise


# %% Calculated partition functions (either from energy levels, or ab initio)


def _get_cachefile_name(ElecState):
    """Get name of cache file for calculated rovibrational energies.

    Basically store it alongside the jsonfile of the ElecState with:

        [jsonfile]_[molecule_name]_[isotope]_[electronic state]_levels.h5
    """
    molecule = ElecState.name
    isotope = ElecState.iso
    state = ElecState.state
    jsonfile = ElecState.jsonfile
    filename = "{0}_{1}_iso{2}_{3}_levels.h5".format(
        jsonfile.replace(".json", ""), molecule, isotope, state
    )
    return filename


PartFuncHAPI = PartFuncTIPS  # old name; for compatibility


class PartFunc_Dunham(RovibParFuncCalculator):
    """Calculate partition functions from spectroscopic constants, if
    molecule data is available in RADIS. If using the
    :ref:`default spectroscopic constants <label_db_spectroscopic_constants>`,
    make sure you know what reference data is being used in RADIS!
    Energies are calculated with the :py:meth:`~radis.db.classes.ElectronicState.Erovib`
    method of :py:data:`~radis.db.molecules.Molecules`

    Parameters
    ----------
    electronic_state: :class:`~radis.db.classes.ElectronicState`
        an :class:`~radis.db.classes.ElectronicState` object, which is
        defined in RADIS molecule database and contains spectroscopic data
    vmax: int, or ``None``
        maximum vibrational quantum to calculate with Dunham expansion.
        If None, the molecule one is taken.
        If None still, all levels are calculated up to the molecule dissociation
        energy
    vmax_morse: int, or ``None``
        maximum vibrational quantum to calculate with Morse potential.
        If None, the molecule one is taken. Use ``0`` or ``-1`` not to calculate
        with Morse potential
    Jmax: int, or ``None``
        maximum rotational quantum. If None, the molecule one is taken.
        If None, all levels are calculated up to the molecule dissociation
        energy

    Other Parameters
    ----------------
    use_cached: ``True``, ``False``, or ``'regen'``, ``'force'``
        if ``True``, use (and generate if doesnt exist) a ``.h5`` file.
        If ``'regen'``, regenerate cache file. If ``'force'``, raise an error
        if file doesnt exist. Default ``True``
    calc_Evib_per_mode: boolean
        if ``True``, calculate energies of each vibration mode (so far only
        implemented for CO2 with Evib1, Evib2, Evib3 but shall be generalized
        to all molecules)
    calc_Evib_harmonic_anharmonic: boolean
        if ``True``, calculate and store separately harmonic and anharmonic parts
        of the vibrational energy. This is needed to calculate Treanor distributions
        ( ~ Evib_harmonic / Tvib  - Evib_anharmonic / Trot )
    group_energy_modes_in_2T_model: dict
        (experimental in neq 0.9.21) for polyatomic molecules (> 1 vibration mode),
        how to regroup energy modes when working with 2T models. For instance,
        for CO2, (Evib, Erot) could as well as defined with::

            ['Evib1', 'Evib2', 'Evib3'],['Erot']
            or
            ['Evib3'],['Evib1', 'Evib2', 'Erot']

        depending on which levels are supposed to interact the most
    mode: 'full summation', 'tabulation'
        calculation mode. ``'tabulation'`` is much faster but not all possible
        distributions are implemented. Default ``'full-summation'``

    Examples
    --------
    Calculate partition function of CO using default spectroscopic constants::

        from radis.db.molecules import Molecules
        from radis.levels.partfunc import PartFunc_Dunham

        isotope = 1
        electronic_state = "X"
        S = Molecules["CO"][isotope][electronic_state]

        # Equilibrium partition functions :
        Qf = PartFunc_Dunham(S)
        print(Qf.at(T=3000))  # K

        # Nonequilibrium partition functions :
        print(Qf.at_noneq(Tvib=2000, Trot=1000))  # K

    .. minigallery:: radis.levels.partfunc.PartFunc_Dunham
        :add-heading:

    Notes
    -----
    Validity:

    So far, RADIS energy levels are only calculated from Dunham's expansion.
    Above a certain vibrational level a Morse Potential may be used. See how
    the molecule is defined in :mod:`~radis.db.molecules`

    See Also
    --------

    :mod:`~radis.db.molecules`,
    :py:meth:`~radis.db.classes.ElectronicState.Erovib`,
    :py:data:`~radis.db.molecules.Molecules`

    """

    def __init__(
        self,
        electronic_state,
        vmax=None,
        vmax_morse=None,
        Jmax=None,
        spectroscopic_constants="default",
        use_cached=True,
        verbose=True,
        calc_Evib_per_mode=True,
        calc_Evib_harmonic_anharmonic=True,
        group_energy_modes_in_2T_model={"CO2": (["Evib1", "Evib2", "Evib3"], ["Erot"])},
        mode="full summation",
    ):  # , ZPE=None):
        # TODO: find a way to initialize calc_Evib_per_mode or calc_Evib_harmonic_anharmonic from
        # the SpectrumFactory on Spectrum generation time...
        # Maybe recompute the cache file if needed?

        # TODO @dev: refactor : move group_energy_modes_in_2T_model in radis/config.json ?

        # WIP @erwan: spectroscopic_constants: NOT USED for the moment
        # parameter to change the constants
        # used to calculate the electronic_state energies. An alternative strategy
        # be to adjust the constants in electronic_state directly.

        # %% Init
        super(PartFunc_Dunham, self).__init__(
            electronic_state=electronic_state, mode=mode, verbose=verbose
        )

        # Check inputs ('return' is not mentioned in signature. it will just return
        # after cache name is given)
        assert use_cached in [True, False, "regen", "force", "return"]

        ElecState = self.ElecState  # molecule
        molecule = ElecState.name  # will be stored in cache file metadata
        isotope = ElecState.iso  # will be stored in cache file metadata
        if vmax is None:
            vmax = ElecState.vmax
        if vmax_morse is None:
            vmax_morse = ElecState.vmax_morse
        if Jmax is None:
            Jmax = ElecState.Jmax

        # How to lump energy modes
        if molecule in group_energy_modes_in_2T_model:
            group_energy_modes = group_energy_modes_in_2T_model[molecule]

        # Store
        self.vmax = vmax
        self.vmax_morse = vmax_morse
        self.Jmax = Jmax
        self.use_cached = use_cached
        self.group_energy_modes_in_2T_model = group_energy_modes_in_2T_model

        # Get variables to store in metadata  (after default values have been set)
        # ... this ensures that cache files generated with different metadata
        # ... will not be used (in particular: different rovib_constants)
        metadata = {
            "molecule": molecule,
            "isotope": isotope,
            "rovib_constants": ElecState.rovib_constants,
            "vmax": vmax,
            "vmax_morse": vmax_morse,
            "Jmax": Jmax,
            "group_energy_modes_in_2T_model": group_energy_modes_in_2T_model,
            "calc_Evib_harmonic_anharmonic": calc_Evib_harmonic_anharmonic,
            "calc_Evib_per_mode": calc_Evib_per_mode,
        }
        if molecule in group_energy_modes_in_2T_model:
            metadata.update({"group_energy_modes": group_energy_modes})

        # get cache file path

        # Function of use_cached value:
        # ... if True, use (and generate if doesnt exist) cache file.
        # ... if 'regen', regenerate cache file. If 'force', raise an error
        # ... if file doesnt exist.
        # If file is deprecated, regenerate it unless 'force' was used

        cachefile = _get_cachefile_name(ElecState)
        self.cachefile = cachefile

        # If return, return after cachefile generated (used for tests)
        if use_cached == "return":
            return

        self.df = load_h5_cache_file(
            cachefile,
            use_cached,
            valid_if_metadata_is=metadata,
            relevant_if_metadata_above={},
            relevant_if_metadata_below={},
            current_version=radis.__version__,
            last_compatible_version=radis.config["OLDEST_COMPATIBLE_VERSION"],
            verbose=verbose,
        )

        # Get levels
        if molecule in HITRAN_CLASS1 + HITRAN_CLASS2 + HITRAN_CLASS3:
            nb_vib_modes = 1
        elif molecule in HITRAN_CLASS5 + HITRAN_CLASS6:
            nb_vib_modes = 3
        else:
            raise NotImplementedError
        self.nb_vib_modes = nb_vib_modes

        # Build energy levels if needed
        if self.df is None:
            # Build energy levels
            if verbose:
                print(
                    "Calculating energy levels with Dunham expansion for {0}".format(
                        ElecState.get_fullname()
                    )
                )
                if not use_cached:
                    print(
                        "Tip: set ``use_cached=True`` next time not to recompute levels every time"
                    )
            if molecule in HITRAN_CLASS1:
                self.build_energy_levels_class1()
            elif molecule in HITRAN_CLASS5:  # CO2
                self.build_energy_levels_class5(
                    calc_Evib_per_mode=calc_Evib_per_mode,
                    calc_Evib_harmonic_anharmonic=calc_Evib_harmonic_anharmonic,
                    group_energy_modes_in_2T_model=group_energy_modes,
                )
            else:
                raise NotImplementedError

        # save files if use_cached
        if use_cached and not exists(cachefile):
            save_to_hdf(
                self.df,
                cachefile,
                metadata=metadata,
                version=radis.__version__,
                key="df",
                overwrite=True,
                verbose=verbose,
            )

        # Add extra columns (note that this is not saved to disk)
        self._add_extra()

        return

    def _add_extra(self):

        df = self.df
        molecule = self.molecule

        if molecule in ["CO2"]:
            # Add harmonic / anharmonic energies
            if (
                all_in(
                    ["Evib1_a", "Evib1_h", "Evib2_a", "Evib2_h", "Evib3_a", "Evib3_h"],
                    df,
                )
                and not "Evib_h" in df
                and not "Evib_a" in df
            ):

                group_energy_modes = self.group_energy_modes_in_2T_model[molecule]

                # Add total harmonic / Anharmonic energies as the sum of each mode
                # (used in 2 T temperature models)

                if group_energy_modes == (["Evib1", "Evib2", "Evib3"], ["Erot"]):
                    df["Evib_h"] = df.Evib1_h + df.Evib2_h + df.Evib3_h
                    df["Evib_a"] = df.Evib1_a + df.Evib2_a + df.Evib3_a

                elif group_energy_modes == (["Evib3"], ["Evib1", "Evib2", "Erot"]):
                    if not all_in(["Evib1", "Evib2", "Evib3"], df):
                        raise KeyError(
                            "You need Evib1, Evib2, Evib3 calculated separately. "
                            + "Use calc_Evib_per_mode=True"
                        )
                    df["Evib_h"] = df.Evib3_h
                    df["Evib_a"] = df.Evib3_a

                else:
                    raise NotImplementedError(
                        "group_energy_modes_in_2T_model: {0}".format(group_energy_modes)
                    )

    def build_energy_levels_class1(self):  # , ZPE=0):
        r"""in the case where only Ediss is given. Deal with vmax, Jmax later.

        Applies to molecules in :data:`~radis.api.hitranapi.HITRAN_CLASS1`

        Returns
        -------

        None:
            but the Pandas DataFrame `self.df` is updated with parameters:

            - ``g`` : degeneracy
            - ``E`` : energy level
            - ``Evib`` : vibrational energy
            - ``Erot`` : rotational energy
            - ``viblvl`` : vibrational level name
        """

        vib_lvl_name = vib_lvl_name_hitran_class1

        ElecState = self.ElecState  # molecule

        Ediss = ElecState.Ediss  # cm-1

        vmax = self.vmax  # max vibrational level to calculate with Dunham
        # max vibrational level to calculate with Morse potential
        vmax_morse = self.vmax_morse
        Jmax = self.Jmax  # max rotational level to calculate

        # First get max vib level  (to be calculated with Dunham expansion)
        if vmax is None:
            vmax = 0
            while ElecState.Erovib(vmax, J=0, remove_ZPE=True) < Ediss:
                vmax += 1
        if Jmax is None:
            Jmax = np.nan  # no limit
        Jmaxcalc = 0  # max calculated. just for info

        # Calculate lower rovibrational levels with Dunham expansion
        # ------------
        # format: (v, j, E, Evib)   # g, Erot calculated at the end (faster)
        levels = []
        Evib = ElecState.Erovib(v=0, J=0, remove_ZPE=True)  # no Zero-point-energy

        # ... Loop over vibrational levels:
        for v in range(0, vmax + 1):
            Evib = ElecState.Erovib(v, J=0, remove_ZPE=True)  # no Zero-point-energy
            if __debug__:
                printdbg(
                    "Calculating Evib for "
                    + "v={0}: {1:.2f}cm-1 (Dunham expansion)".format(v, Evib)
                )
            # ... Rotational loop
            E = Evib
            J = 0
            while 0 <= E < Ediss and not J > Jmax:  # (passes if Jmax is nan)
                # store rovib level
                levels.append([v, J, E, Evib])
                # calculate new one:
                J += 1
                E = ElecState.Erovib(v, J, remove_ZPE=True)  # no Zero-point-energy
            Jmaxcalc = max(Jmaxcalc, J)

        # If defined, calculate upper rovibrational levels with Morse potential
        # ----------
        if vmax_morse is not None:

            # vibrational energy of the last vibrational level for which Dunham expansion is valid
            Evib_last = ElecState.Erovib(vmax, 0)
            # difference of the 2 last vib level for which Dunham expansion is valid
            delta_E_last = Evib_last - ElecState.Erovib(vmax - 1, 0)

            v_inc = ElecState.get_Morse_inc()

            # ... Start loop on all Morse potential levels
            Evib = Evib_last
            for v in range(vmax + 1, vmax_morse + 1):
                delta_E = delta_E_last - (v + 1 - vmax) * v_inc
                Evib = Evib + delta_E
                if Evib > Ediss:
                    warn(
                        "Energy above dissociation threshold: {0}>{1}".format(
                            Evib, Ediss
                        )
                    )
                if __debug__:
                    printdbg(
                        "Calculating Evib for "
                        + "v={0}: {1:.2f}cm-1 (Morse Potential)".format(v, Evib)
                    )
                # ... Rotational loop
                J = 0
                E = Evib
                # (passes if Jmax is nan)
                while 0 <= E < Ediss and not J >= Jmax:
                    # store rovib level
                    levels.append([v, J, E, Evib])
                    # calculate new one:
                    J += 1
                    # no Zero-point-energy
                    Erot = ElecState.Erovib(0, J, remove_ZPE=True)
                    E = Evib + Erot
                Jmaxcalc = max(Jmaxcalc, J)

        df = pd.DataFrame(levels, columns=["v", "j", "E", "Evib"])

        # Store vibrational level name
        df["viblvl"] = vib_lvl_name(df.v)

        # Calculate degeneracies
        df["gj"] = 2 * df.j + 1
        df["gvib"] = 1  # energy base is assumed to be rovibrational complete
        gs = self.gs(ElecState)

        if isinstance(gs, tuple):
            raise NotImplementedError(
                "Different degeneracies (symmetric/antisymmetric)"
                + " not implemented for molecule {0} isotope {1}".format(
                    ElecState.id, ElecState.iso
                )
            )
            # (for devs: to implement that efficiently see CO2 (_class5))

        gi = self.gi(ElecState)
        df["grot"] = gs * gi * df.gj
        df["Erot"] = df.E - df.Evib

        if self.verbose:
            print("Database generated up to v={0}, J={1}".format(v, Jmaxcalc))

        self.df = df

    def build_energy_levels_class5(
        self,
        calc_Evib_per_mode=True,
        calc_Evib_harmonic_anharmonic=False,
        group_energy_modes_in_2T_model=(["Evib1", "Evib2", "Evib2"], ["Erot"]),
    ):
        r"""in the case where only Ediss is given. Deal with vmax, Jmax later.

        :data:`~radis.api.hitranapi.HITRAN_CLASS5` = ['CO2']
        # Linear triatomic with large Fermi resonance

        Parameters
        ----------
        calc_Evib_per_mode: boolean
            if ``True``, calculates Evib1, Evib2, Evib3

        Other Parameters
        ----------------
        calc_Evib_harmonic_anharmonic: boolean
            if ``True``, calculates Evib_h and Evib_a, harmonic and non harmonic
            components, to be used in Treanor distributions
        group_energy_modes_in_2T_model: dict
            (experimental in neq 0.9.21) for polyatomic molecules (> 1 vibration mode),
            how to regroup energy modes when working with 2T models. For instance,
            for CO2, (Evib, Erot) could as well as defined with::

                ['Evib1', 'Evib2', 'Evib3'],['Erot']

            or::

                ['Evib3'],['Evib1', 'Evib2', 'Erot']

            depending on which levels are supposed to interact the most

        Returns
        -------
        None:
            but the Pandas DataFrame `self.df` is updated with parameters:

            - ``g`` : degeneracy
            - ``E`` : energy level
            - ``Evib`` : vibrational energy
            - ``Erot`` : rotational energy
            - ``viblvl`` : vibrational level name
        """

        vib_lvl_name = vib_lvl_name_hitran_class5

        ElecState = self.ElecState
        Ediss = ElecState.Ediss  # cm-1

        vmax = self.vmax
        vmax_morse = self.vmax_morse
        Jmax = self.Jmax
        if vmax_morse is not None:
            raise NotImplementedError

        if vmax is None:
            vmax = ElecState.vmax
        if Jmax is None:
            Jmax = ElecState.Jmax

        #        if vmax is None: vmax = 100    # just to prevent infinite loop. Ediss should be limited
        if Jmax is None:
            Jmax = 1000

        # First get max levels
        v1max = 0
        v2max = 0
        v3max = 0
        while ElecState.Erovib(v1max, 0, 0, 0, J=0, remove_ZPE=True) < Ediss:
            v1max += 1
        while ElecState.Erovib(0, v2max, v2max, 0, J=0, remove_ZPE=True) < Ediss:
            v2max += 1
        while ElecState.Erovib(0, 0, 0, v3max, J=0, remove_ZPE=True) < Ediss:
            v3max += 1
        if vmax is not None:
            v1max = min(vmax, v1max)
            v2max = min(vmax, v2max)
            v3max = min(vmax, v3max)

        # Then fill mixed modes levels

        # %%

        levels = []  # vibrational levels
        pb = ProgressBar(v1max + 1, active=True)
        Jmax_calc = 0
        for v1 in range(v1max + 1):
            pb.update(v1)
            for v2 in range(v2max + 1):
                for l2 in [v2]:
                    # Calculation with HITRAN spectroscopic convention: v2=l2
                    # instead of l2 = [v2::v2+1::2]
                    # It follows than gvib = v2+1
                    # This is added later
                    for v3 in range(v3max + 1):
                        """
                        Spectroscopic rule
                        ------------------
                        J>=l2: as in the symmetric rotor
                        ------------------
                        """
                        viblvl = vib_lvl_name(v1, v2, l2, v3)

                        # First calculate vibrational energy only, for 2-T models
                        if not calc_Evib_harmonic_anharmonic:

                            # recompute
                            # Note Evib = Evib1 + Evib2 + Evib3 should be true
                            # if there is no coupling. General case here: we
                            # recompute each separately
                            if calc_Evib_per_mode:
                                Evib1 = ElecState.Erovib(v1, 0, 0, 0, 0)
                                Evib2 = ElecState.Erovib(0, v2, l2, 0, 0)
                                Evib3 = ElecState.Erovib(0, 0, 0, v3, 0)
                                Evib123 = ElecState.Erovib(v1, v2, l2, v3, 0)
                                # note: the latest could be different from Evib1+Evib2+Evib3
                                # if perturbations are taken into account
                            else:
                                Evib123 = ElecState.Erovib(v1, v2, l2, v3, 0)

                        else:  # calc_Evib_harmonic_anharmonic

                            pass
                            # NotImplemented. Or rather, (G1_h, G1_a), (G2_h, G2_a), (G3_h, G3_a)
                            # is returned anyway by the call to ElecState.Ehaj

                        for J in range(l2, Jmax + 1):
                            #                            roviblvl = '({0}, {1}, {2}, {3}, {4})'.format(v1, v2, l2, v3, J)

                            # Energy level rules

                            # Note: for most isotopes only half the levels exist.
                            # because the other have a gs=0 degeneracy.

                            # Here we still calculate the energy for all
                            # levels because for some reason CDSD features
                            # transitions with some of these levels, and
                            # it creates KeyErrors if we dont
                            # (can there be non null 'forbidden' levels as
                            # there are forbidden transitions?).
                            # Anyway, their linestrengths are very small,
                            # and they wont be accounted for in the
                            # partition function because of gs=0

                            if not calc_Evib_harmonic_anharmonic:

                                # harmonic, corrected
                                Etot = ElecState.Erovib(v1, v2, l2, v3, J)
                                if Etot > Ediss:
                                    break
                                # recompute
                                # Note Evib = Evib1 + Evib2 + Evib3 should be true
                                # if there is no coupling. General case here: we
                                # recompute each separately
                                if calc_Evib_per_mode:
                                    # the following has been precomputed before the
                                    # rotational loop
                                    #                                    Evib1 = ElecState.Erovib(v1, 0, 0, 0, 0)
                                    #                                    Evib2 = ElecState.Erovib(0, v2, l2, 0, 0)
                                    #                                    Evib3 = ElecState.Erovib(0, 0, 0, v3, 0)
                                    #                                    Evib = ElecState.Erovib(v1, v2, l2, v3, 0)

                                    # redefining 'columns' names at each iteration, but
                                    # there is less risk to invert names and data
                                    columns = [
                                        "v1",
                                        "v2",
                                        "l2",
                                        "v3",
                                        "j",
                                        "viblvl",
                                        "E",
                                        "Evib123",
                                        "Evib1",
                                        "Evib2",
                                        "Evib3",
                                    ]
                                    levels.append(
                                        [
                                            v1,
                                            v2,
                                            l2,
                                            v3,
                                            J,
                                            viblvl,
                                            Etot,
                                            Evib123,
                                            Evib1,
                                            Evib2,
                                            Evib3,
                                        ]
                                    )
                                else:
                                    # the following has been precomputed before the
                                    # rotational loop
                                    #                                    Evib = ElecState.Erovib(v1, v2, l2, v3, 0)

                                    # redefining 'columns' names at each iteration, but
                                    # there is less risk to invert names and data
                                    columns = [
                                        "v1",
                                        "v2",
                                        "l2",
                                        "v3",
                                        "j",
                                        "viblvl",
                                        "E",
                                        "Evib123",
                                    ]
                                    levels.append(
                                        [v1, v2, l2, v3, J, viblvl, Etot, Evib123]
                                    )

                            else:  # calc_Evib_harmonic_anharmonic

                                if calc_Evib_per_mode:
                                    (
                                        (G1_h, G1_a),
                                        (G2_h, G2_a),
                                        (G3_h, G3_a),
                                        FJ,
                                    ) = ElecState.Ehaj(v1, v2, l2, v3, J)
                                    Evib1 = G1_h + G1_a
                                    Evib2 = G2_h + G2_a
                                    Evib3 = G3_h + G3_a
                                    Evib123 = Evib1 + Evib2 + Evib3
                                    Etot = Evib123 + FJ
                                    if Etot > Ediss:
                                        break
                                    # recompute
                                    # store

                                    # redefining 'columns' names at each iteration, but
                                    # there is less risk to invert names and data
                                    columns = [
                                        "v1",
                                        "v2",
                                        "l2",
                                        "v3",
                                        "j",
                                        "viblvl",
                                        "E",
                                        "Evib123",
                                        "Evib1",
                                        "Evib2",
                                        "Evib3",
                                        "Evib1_h",
                                        "Evib1_a",
                                        "Evib2_h",
                                        "Evib2_a",
                                        "Evib3_h",
                                        "Evib3_a",
                                    ]
                                    levels.append(
                                        [
                                            v1,
                                            v2,
                                            l2,
                                            v3,
                                            J,
                                            viblvl,
                                            Etot,
                                            Evib123,
                                            Evib1,
                                            Evib2,
                                            Evib3,
                                            G1_h,
                                            G1_a,
                                            G2_h,
                                            G2_a,
                                            G3_h,
                                            G3_a,
                                        ]
                                    )
                                else:
                                    raise NotImplementedError

                            Jmax_calc = max(J, Jmax_calc)
        pb.done()

        df = pd.DataFrame(levels, columns=columns)

        # Calculate missing energies
        # --------------------------

        df["Erot"] = df.E - df.Evib123

        if group_energy_modes_in_2T_model == (["Evib1", "Evib2", "Evib3"], ["Erot"]):
            df["Evib"] = df.Evib123  # + Evib2 + Evib3
            df["Erot"] = df.Erot

        elif group_energy_modes_in_2T_model == (["Evib3"], ["Evib1", "Evib2", "Erot"]):
            if not all_in(["Evib1", "Evib2", "Evib3"], df):
                raise KeyError(
                    "You need Evib1, Evib2, Evib3 calculated separately. "
                    + "Use calc_Evib_per_mode=True"
                )
            df["Evib"] = df.Evib3
            df["Erot"] = df.Evib1 + df.Evib2 + df.Erot

        else:
            raise NotImplementedError(
                "group_energy_modes_in_2T_model: {0}".format(
                    group_energy_modes_in_2T_model
                )
            )
        #         # same as above but harder to read
        #        Evib_columns, Erot_columns = group_energy_modes_in_2T_model
        #        df['Evib20'] = pd.concat([df[k] for k in Evib_columns])
        #        df['Erot2'] = pd.concat([df[k] for k in Erot_columns])
        assert np.allclose(df.Evib + df.Erot, df.E)

        # Get Degeneracies
        # ----------------

        # ... remember: HITRAN convention with v2=l2 -> gvib degeneracy is "v2+1"
        # TODO: make a function call to radis.db.degeneracies.gvib ?
        df["gvib"] = df.v2 + 1
        df["gj"] = 2 * df.j + 1

        gi = self.gi(ElecState)
        gs = self.gs(ElecState)

        # %%
        def is_symmetric(v1, v2, l2, v3):
            """Returns whether a CO2 ``v1v2'l2'v3`` vibrational level is
            symmetric.

            Notes
            -----

            Spectroscopic rules for CO2:

            - Because ground state CO2 is symmetric, and oxygen :math:`^{16}O`
             has no spin, for the CO2 1st isotope (626) the negative
             rotational levels* are missing

            *We remind than rotational levels are called positive/negative,
            if the total eigenfunction (including all modes) is
            unchanged/changed, respectively, by reflection of all nuclei
            and electrons at the origin.

            Hence:

            - symmetric vibrational levels -> only even j numbers
            - asymmetric vibrational levels -> only odd j numbers

            If oxygen atoms are different isotopologues (ex: CO2 628), then all
            levels exist

            References
            ----------

            See Section (2.3) in ``The CO2 Laser by Witteman 1987,
            ISBN 3540477446, 9783540477440``

            Examples
            --------

            Output of ``is_symmetric(v1, v2, l2, v3)`` for different levels::

                >>> is_symmetric(1,0,0,0)
                True
                >>> is_symmetric(2,0,0,0)
                True
                >>> is_symmetric(0,2,0,0)
                True
                >>> is_symmetric(0,2,2,0)
                True
                >>> is_symmetric(0,0,0,2)
                True
                >>> is_symmetric(0,0,0,1)
                False
                >>> is_symmetric(0,2,0,1)
                False
                >>> is_symmetric(1,0,0,1)
                False
                >>> is_symmetric(0,3,1,0)
                False
            """
            sym = (-1) ** v2 * (-1) ** v3
            return sym == 1

        def is_even(J):
            """Return whether J is an even or odd rotational level."""
            #            return not bool(J % 2)   # works for float only
            return 1 - np.mod(J, 2)

        if isinstance(gs, tuple):
            # Different degeneracy for symmetric and antisymmetric rotational
            # levels for this isotope. We need the symmetry of rotational levels!

            # Get symmetry of rotational levels

            # Spectroscopic rules
            # -------------------
            # for a spin-symmetric isotope (CO2 626):
            # - if vibrational level is asymmetric, even (+) rotational
            #   levels do not exist
            # - if vibrational level is symmetric, odd (-) rotational
            #   levels do not exist
            # -------------------

            # Get vibrational symmetry: True if Symmetric, False if Antisymmetric
            viblvl_is_symmetric = is_symmetric(
                df.v1.values, df.v2.values, df.l2.values, df.v3.values
            )
            # Get rotational parity: True if even, False if odd
            rotlvl_is_even = is_even(df.j.values)

            # Derive rotational symmetry: True if Symmetric
            rotlvl_is_symmetric = viblvl_is_symmetric == rotlvl_is_even

            # Get state-dependant parity for symmetric and antisymmetric levels
            gs_s, gs_a = gs
            gs = np.ones_like(df.j.values) * gs_a
            gs[rotlvl_is_symmetric] = gs_s

        # Calculate rotational degeneracy
        df["grot"] = gs * gi * df.gj

        #        # just checking:
        #        if not (df.Evib == (df.Evib1 + df.Evib2 + df.Evib3)).all():
        #            raise ValueError('Vib energy is not the sum of all mode energies')
        # not necessarily the case anymore: could be Evib, Erot = Evib3, Evib1+Evib2+Erot

        if self.verbose:
            print(
                "Database generated up to v1={0}, v2={1}, v3={2}, J={3}".format(
                    v1max, v2max, v3max, Jmax_calc
                )
            )

        self.df = df

    def gs(self, ElecState):  # , viblvl):
        r"""Get state specific rotational degeneracy.

        Parameters
        ----------
        ElecState
            an ElectronicState, that contains molecule id and isotope number

        See Also
        --------

        :func:`~radis.db.degeneracies.gs`
        """

        from radis.db.degeneracies import gs

        M, I = ElecState.id, ElecState.iso
        return gs(M, I)

    def gi(self, ElecState):
        r"""Get state independent rotational degeneracy. Typically depends on
        the isotope.

        See Also
        --------

        :func:`~radis.db.degeneracies.gi`
        """
        from radis.db.degeneracies import gi

        M, I = ElecState.id, ElecState.iso
        return gi(M, I)


# %% Test
if __name__ == "__main__":

    from radis.test.levels.test_partfunc import _run_testcases

    print("Testing parfunc: {0}".format(_run_testcases()))



# -*- coding: utf-8 -*-
"""
Created on Wed Aug 29 10:35:24 2018

@author: erwan
"""

# from time import perf_counter

import numpy as np
import pytest

# from radis import get_residual
# from radis.lbl.factory import SpectrumFactory
from radis.misc.arrays import (  # add_at,
    arange_len,
    autoturn,
    bining,
    calc_diff,
    centered_diff,
    find_first,
    find_nearest,
    first_nonnan_index,
    is_sorted,
    is_sorted_backward,
    last_nonnan_index,
    logspace,
    non_zero_values_around,
)

# from radis.test.utils import setup_test_line_databases


@pytest.mark.fast
def test_arange_len(*args, **kwargs):

    # Positive arrays
    wmin, wmax, wstep = (380, 700, 0.1)
    assert arange_len(wmin, wmax, wstep) == len(np.arange(wmin, wmax, wstep))
    wmin, wmax, wstep = (380, 700, 0.31)
    assert arange_len(wmin, wmax, wstep) == len(np.arange(wmin, wmax, wstep))

    # Negative arrays
    wmin, wmax, wstep = (-100, 0, 0.1)
    assert arange_len(wmin, wmax, wstep) == len(np.arange(wmin, wmax, wstep))

    wmin, wmax, wstep = (-100, 0, 0.31)
    assert arange_len(wmin, wmax, wstep) == len(np.arange(wmin, wmax, wstep))

    # Centered arrays
    wmin, wmax, wstep = (-100, 100, 0.1)
    assert arange_len(wmin, wmax, wstep) == len(np.arange(wmin, wmax, wstep))
    wmin, wmax, wstep = (-100, 100, 0.31)
    assert arange_len(wmin, wmax, wstep) == len(np.arange(wmin, wmax, wstep))

    wmin, wmax = (-100, 100)
    for wstep in np.random.rand(100):
        assert arange_len(wmin, wmax, wstep) == len(np.arange(wmin, wmax, wstep))


@pytest.mark.fast
def test_is_sorted(*args, **kwargs):

    a = np.arange(10)

    assert is_sorted(a)
    assert is_sorted_backward(a[::-1])

    assert not is_sorted_backward(a)
    assert not is_sorted(a[::-1])


@pytest.mark.fast
def test_nonnan_index(*args, **kwargs):

    a = np.arange(1000) * 0.2

    assert first_nonnan_index(a) == 0  # is None
    assert last_nonnan_index(a) == 999  # is None

    a[:10] = np.nan
    a[-60:] = np.nan

    assert first_nonnan_index(a) == 10
    assert last_nonnan_index(a) == 939

    a[:] = np.nan

    assert first_nonnan_index(a) == None
    assert last_nonnan_index(a) == None  # len(a)-1


@pytest.mark.fast
def test_find_first(*args, **kwargs):

    a = np.arange(10)
    assert find_first(a, -1) == 0
    assert find_first(a, 0) == 1
    assert find_first(a, 5) == 6
    assert find_first(a, 8) == 9
    assert find_first(a, 9) == 0
    assert find_first(a, 20) == 0

    assert not find_first(a, -1) == -1
    assert not find_first(a, 0) == 0
    assert not find_first(a, 5) == 5
    assert not find_first(a, 10) == 10


@pytest.mark.fast
def test_bining(*args, **kwargs):

    a = np.arange(20).reshape(4, 5)
    assert (bining(a) == np.array([2, 7, 12, 17])).all()
    assert (bining(a, ymin=1) == np.array([2.5, 7.5, 12.5, 17.5])).all()
    assert (bining(a, ymax=3) == np.array([1, 6, 11, 16])).all()
    assert (bining(a, ymin=1, ymax=3) == np.array([1.5, 6.5, 11.5, 16.5])).all()


@pytest.mark.fast
def test_calc_diff(*args, **kwargs):
    t1 = np.arange(5)
    t2 = np.arange(5)

    v1 = np.array([10, 12, 14, 16, 18])
    v2 = np.array([10, 12, 14, 16, 18])

    t_res1, v_res1 = calc_diff(t1, v1, t2, v2)
    t_res2, v_res2 = calc_diff(t1, v1, t1[::-1], v2)
    t_res3, v_res3 = calc_diff(t2[::-1], v1, t2, v2)

    assert (t_res1 == np.array([1, 2, 3])).all()
    assert (v_res1 == np.array([0, 0, 0])).all()

    assert (t_res2 == np.array([1, 2, 3])).all()
    assert (v_res2 == np.array([-4, 0, 4])).all()

    assert (t_res3 == np.array([1, 2, 3])).all()
    assert (v_res3 == np.array([4, 0, -4])).all()


@pytest.mark.fast
def test_autoturn(*args, **kwargs):
    dat = np.arange(20)
    dat.resize(2, 10)
    dat_rot = np.transpose(dat)

    assert (autoturn(dat, key=0) == dat).all()
    assert (autoturn(dat, key=1) == dat_rot).all()
    assert (autoturn(dat) == dat).all()


@pytest.mark.fast
def test_centered_diff(*args, **kwargs):
    a = np.arange(10)
    ones = np.ones_like(a)
    zeros = np.zeros_like(a)

    assert (centered_diff(a) == ones).all()
    assert not (centered_diff(a) == zeros).all()
    assert (centered_diff(ones) == zeros).all()
    assert len(centered_diff(a)) == len(a)


@pytest.mark.fast
def test_logspace(*args, **kwargs):
    dat1 = logspace(1, 100, 10)
    dat2 = logspace(17, 250, 37)
    dat3 = logspace(5, 19, 3)

    dats = [dat1, dat2, dat3]

    assert (dats[0][0] - 1) <= 1e-6 and (dats[0][9] - 100) <= 1e-6
    assert (dats[1][0] - 17) <= 1e-6 and (dats[1][36] - 250) <= 1e-6
    assert (dats[2][0] - 5) <= 1e-6 and (dats[2][2] - 19) <= 1e-6

    for dat in dats:
        for i in range(2, len(dat)):
            assert (dat[i] / dat[i - 1] - dat[i - 1] / dat[i - 2]) <= 1e-6


@pytest.mark.fast
def test_find_nearest(*args, **kwargs):
    a = np.arange(10)
    b = np.ones(5)

    x1, y1 = find_nearest(a, a, True)
    assert (x1 == a).all()
    assert len(y1) == len(a)
    assert (y1 == np.array([True for i in a])).all()

    x2, y2 = find_nearest(a, b, True)
    assert (x2 == b).all()
    assert len(x2) == len(b)
    assert len(y2) == len(a)
    assert (y2 == np.array([i == 1 for i in a])).all()

    x3, y3 = find_nearest(np.array([10]), np.array([-10, 0, 10]), True)
    assert (x3 == np.array([10, 10, 10])).all()
    assert len(y3) == 1
    assert y3 == [True]

    x4, y4 = find_nearest(np.array([1, 4, 3]), np.array([]), True)
    assert (x4 == np.array([])).all()
    assert len(y4) == 3
    assert (y4 == [False for _ in range(3)]).all()

    assert (find_nearest(a, a[::-1]) == a[::-1]).all()
    assert (
        find_nearest(np.array([1.5, 2.0, 2.5, 3.0]), np.array([-10, 10, 2.25]))
        == np.array([1.5, 3.0, 2.0])
    ).all()

    assert (find_nearest(np.array([1, 3]), np.array([2])) == np.array([1])).all()
    assert (find_nearest(np.array([3, 1]), np.array([2])) == np.array([3])).all()


# @pytest.mark.fast
# def test_cython_add_at(*args, **kwargs):
##    """
##    Compare the workings of the Cython compiled add_at() function
##    versus the numpy add.at() function with bogus data.
##    """
##
##    # First check if Python was able to import the Cython version of add at:
##    from radis_cython_extensions import add_at as cython_add_at
##
##    assert add_at == cython_add_at
##
##    # Compare output LDM's between the two additions:
##    Nv = 300000
##    NG = 4
##    NL = 16
##    Ni = 1000000
##
##    I = np.random.rand(Ni).astype(np.float32)
##    k = np.random.randint(Nv, size=Ni, dtype=np.int32)
##    l = np.random.randint(NG, size=Ni, dtype=np.int32)
##    m = np.random.randint(NL, size=Ni, dtype=np.int32)
##
##    LDM1 = np.zeros((Nv, NG, NL), dtype=np.float32)
##    t0 = perf_counter()
##    np.add.at(LDM1, (k, l, m), I)
##    print("Numpy add.at(): ", perf_counter() - t0)
##
##    LDM2 = np.zeros((Nv, NG, NL), dtype=np.float32)
##    t0 = perf_counter()
##    cython_add_at(LDM2, k, l, m, I)
##    print("Cython add_at(): ", perf_counter() - t0)
##
##    print("Residual: ", np.sum(np.abs(LDM1 - LDM2)))
##    assert np.allclose(LDM1, LDM2)
# assert True


# def test_cython_add_at_spectra(*args, **kwargs):
##    """
##    Test if the Cython add_at() produces the same spectra as
##    with numpy add.at().
##    """
##
##    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/radis.json if not there
##
##    # Conditions
##    wstep = 0.005
##    wmin = 2100  # cm-1
##    wmax = 2200  # cm-1
##
##    T = 1200  # K
##    p = 0.1  # bar
##
##    sf = SpectrumFactory(
##        wavenum_min=wmin,
##        wavenum_max=wmax,
##        mole_fraction=1,
##        path_length=1,  # doesnt change anything
##        wstep=wstep,
##        pressure=p,
##        isotope="1",
##        verbose=False,
##        warnings={
##            "MissingSelfBroadeningWarning": "ignore",
##            "NegativeEnergiesWarning": "ignore",
##            "HighTemperatureWarning": "ignore",
##            "OutOfRangeLinesWarning": "ignore",
##            "GaussianBroadeningWarning": "ignore",
##            "CollisionalBroadeningWarning": "ignore",
##            "AccuracyWarning": "ignore",
##        },
##    )
##    sf.load_databank("HITRAN-CO-TEST")
##
##    sf.use_cython = False
##    s_numpy = sf.eq_spectrum(Tgas=T, name="numpy")
##    s_numpy.apply_slit(0.5, "nm")
##    assert sf.misc.add_at_used == "numpy"
##
##    sf.use_cython = True
##    s_cython = sf.eq_spectrum(Tgas=T, name="cython")
##    s_cython.apply_slit(0.5, "nm")
##    assert sf.misc.add_at_used == "cython"
##
##    res = get_residual(s_numpy, s_cython, "transmittance")
##    assert res < 2e-4
# assert True


def test_non_zero_values_around(*args, **kwargs):
    """return a boolean array of same size as ``a`` where each position ``i``
    is ``True`` if there are non-zero points less than ``n`` index position
    away from ``a[i]``, and ``False`` if all points in ``a`` are 0 ``n``  index
    position away from ``a[i]``
    """

    a = np.array(
        [0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0], dtype=np.float64
    )

    # n = 1
    b = np.array([0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1])
    out = non_zero_values_around(a, 1)
    print(a, "width 1")
    print(b)
    print(np.array(out, dtype=int))
    assert (out == b).all()

    # n = 2
    b = np.array([1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1])
    out = non_zero_values_around(a, 2)
    print(a, "width 2")
    print(b)
    print(np.array(out, dtype=int))
    assert (out == b).all()

    # n = 2
    a = np.array(
        [0, 1, 1, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1], dtype=np.float64
    )
    b = np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1])
    out = non_zero_values_around(a, 2)
    print(a, "width 2")
    print(b)
    print(np.array(out, dtype=int))
    assert (out == b).all()

    # n = 1
    a = np.array(
        [0, 1, 1, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1], dtype=np.float64
    )
    b = np.array([1, 1, 1, 1, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1])
    out = non_zero_values_around(a, 1)
    print(a, "width 1")
    print(b)
    print(np.array(out, dtype=int))
    assert (out == b).all()

    # Also test non_zero_ranges_in_array, boolean_array_from_ranges
    from radis.misc.arrays import boolean_array_from_ranges, non_zero_ranges_in_array

    b = np.array([0, 0, 1, 1, 0, 1, 0, 1], dtype=bool)
    assert (non_zero_ranges_in_array(b) == np.array([(2, 4), (5, 6), (7, 8)])).all()

    L = np.array([[2, 4], [5, 6], [7, 8]], dtype=np.int64)
    assert (
        boolean_array_from_ranges(L, 8)
        == np.array([0, 0, 1, 1, 0, 1, 0, 1], dtype=bool)
    ).all()


if __name__ == "__main__":

    pytest.main(["test_arrays.py", "-s"])  # -s for showing console output



# -*- coding: utf-8 -*-
"""
Benchmark RADIS performances against a HAPI equilibrium simulation of Methane

Notes
-----

We try to use the same Computation parameters when possible, in particular
the :py:attr:`~radis.lbl.factory.SpectrumFactory.broadening_max_width` of RADIS
is defined in terms of WavenumberWingHW/HWHM in [HAPI]_. This
benchmark case of [HAPI], Table 7, Methane_III, is also featured in the [RADIS-2018] article.



References
----------

[HAPI]_ article, Table 7, case Methane III with SCPF, 50 HWHMs  :

    >>> Calculation time in article: 317 s


Typical results on an XPS 15 laptop here::

    >>> Calculated with HAPI in 157.41s
    >>> Calculated with RADIS in 1.65s


-------------------------------------------------------------------------------

"""

from os.path import dirname, join
from time import time

from hapi import ISO_ID, absorptionCoefficient_Voigt, db_begin, fetch_by_ids, tableList

from radis import Spectrum, SpectrumFactory, plot_diff

if __name__ == "__main__":

    benchmark_line_brd_ratio = 50  # “WavenumberWingHW”/HWHMs
    dnu = 0.01  # step in HAPI Benchmark article
    molecule = "CH4"
    wavenum_min = 0.001
    wavenum_max = 11505
    pressure_bar = 1.01315
    T = 296
    isotopes = [1, 2, 3, 4]

    sf = SpectrumFactory(
        wavenum_min=wavenum_min,
        wavenum_max=wavenum_max,
        isotope=isotopes,  #'all',
        verbose=2,
        wstep=dnu,  # depends on HAPI benchmark.
        cutoff=1e-23,
        truncation=2.865,  # Corresponds to WavenumberWingHW/HWHM=50 in HAPI
        molecule=molecule,
        optimization=None,
    )
    sf.fetch_databank("hitran")

    s = sf.eq_spectrum(Tgas=T, pressure=pressure_bar)
    s.name = "RADIS ({0:.1f}s)".format(s.conditions["calculation_time"])

    # Print our HWHM for comparison (a posteriori)
    print(("HWHM max {0:.2f} cm-1".format(sf.df1.hwhm_voigt.max())))
    print(
        (
            "WavenumberWingHW/HWHM",
            int(sf.params.truncation / (sf.df1.hwhm_voigt.max())),
        )
    )
    assert (
        int(sf.params.truncation / (sf.df1.hwhm_voigt.max()))
    ) == benchmark_line_brd_ratio

    # %% Run HAPI
    print("Now calculating with HAPI")

    # Generate HAPI database locally
    db_begin(join(dirname(__file__), __file__.replace(".py", "_HAPIdata")))
    if not molecule in tableList():  # only if data not downloaded already
        mol_iso_ids = [k for k in ISO_ID if ISO_ID[k][-1] == molecule]
        mol_iso_ids = [k for k in mol_iso_ids if ISO_ID[k][1] in isotopes]
        fetch_by_ids(molecule, mol_iso_ids, wavenum_min, wavenum_max)

    # Calculate with HAPI
    def calc_hapi():
        nu, coef = absorptionCoefficient_Voigt(
            SourceTables=molecule,
            Environment={
                "T": T,
                "p": pressure_bar / 1.01315,
            },  # K  # atm
            GammaL="gamma_self",
            WavenumberStep=dnu,
            HITRAN_units=False,
        )
        return nu, coef

    t0 = time()
    nu, coef = calc_hapi()
    t0 = time() - t0
    print(("Calculated with HAPI in {0:.2f}s".format(t0)))

    s_hapi = Spectrum.from_array(nu, coef, "abscoeff", wunit="cm-1", unit="cm-1")
    s_hapi.name = "HAPI ({0:.1f}s)".format(t0)

    plot_diff(s_hapi, s, "abscoeff")

    print(
        ("Calculated with RADIS in {0:.2f}s".format(s.conditions["calculation_time"]))
    )
    print(("Number of lines in RADIS:", len(sf.df0)))

#    plt.savefig('radis_vs_hapi_test_large_ch4.pdf')



# -*- coding: utf-8 -*-
"""
Created on Mon Nov 20 10:59:49 2017

@author: erwan

Examples
--------

Run all tests::

    pytest       (in command line, in project folder)

Run only fast tests (i.e: tests that a 'fast' label)::

    pytest -m fast


-------------------------------------------------------------------------------


"""

import os
from os.path import basename, exists

import matplotlib.pyplot as plt

from radis.lbl import SpectrumFactory
from radis.misc.printer import printm
from radis.test.utils import setup_test_line_databases
from radis.tools.database import load_spec

fig_prefix = basename(__file__) + ": "

# %% Test


def test_load_spectrum(plot=False, verbose=True, warnings=True, *args, **kwargs):
    """Test load / save

    Fast version: dont save lines / populations, compare spectra only
    """

    setup_test_line_databases()

    temp_file_name = "_test_database_co2_tempfile.spec"
    assert not exists(temp_file_name)

    try:
        sf = SpectrumFactory(
            wavelength_min=4190,
            wavelength_max=4200,
            mole_fraction=400e-6,
            path_length=0.1,  # cm
            isotope=[1],
            cutoff=1e-20,
            verbose=verbose,
        )
        sf.warnings["MissingSelfBroadeningWarning"] = "ignore"
        sf.load_databank("HITRAN-CO2-TEST")
        s1 = sf.eq_spectrum(Tgas=300)
        s1.apply_slit(2, verbose=False)
        s1.update()
        s2 = load_spec(s1.store(temp_file_name, compress=True))
        s2.update()
        if plot:
            fig = plt.figure(fig_prefix + "Calculated vs stored+retrieved")
            s1.plot("absorbance", nfig=fig.number, lw=3, label="calculated")
            s2.plot(
                "absorbance",
                nfig=fig.number,
                color="r",
                label="stored (compressed) and retrieved",
            )
            plt.legend()

        assert s1.compare_with(s2, spectra_only=True, plot=False)

    finally:  # cleaning
        if exists(temp_file_name):
            os.remove(temp_file_name)

    return True


def test_load_lines_pops(plot=False, verbose=True, warnings=True, *args, **kwargs):
    """Test load / save

    Full version: save lines, populations (hundreds of MB for CO2, much less
    for CO), compare everything
    """

    temp_file_name = "_test_database_co_tempfile.spec"
    assert not exists(temp_file_name)

    try:
        sf = SpectrumFactory(
            wavelength_min=4500,
            wavelength_max=4800,
            mole_fraction=400e-6,
            path_length=0.1,  # cm
            isotope=[1, 2],
            cutoff=1e-20,
            verbose=verbose,
        )
        sf.warnings["MissingSelfBroadeningWarning"] = "ignore"
        sf.load_databank("HITRAN-CO-TEST")
        s1 = sf.non_eq_spectrum(Tvib=300, Trot=300)
        s1.apply_slit(2, verbose=False)
        s1.update()
        s2 = load_spec(
            s1.store(
                temp_file_name,
                discard=[],  # we're actually saving lines and
                # populations here! expect hundreds
                # of megabytes for big molecules
                # :line CO2 ... here CO is fine
                # (by default lines and populations
                # are discarded)
                compress=True  # only removes some spectral
                # quantities, cant change population
                # or lines size
            ),
            binary=True,  # uncompress
        )
        s2.update()

        assert s1.compare_with(s2, spectra_only=False, plot=False)

        # Test with json-tricks directly
        # Does not work yet
    #        from json_tricks import dumps, loads
    #        s2b = loads(dumps(s1))
    #        assert s1.compare_with(s2b, spectra_only=False, plot=False)

    finally:  # cleaning
        if exists(temp_file_name):
            os.remove(temp_file_name)

    return True


def _run_testcases(plot=False, verbose=True, warnings=True, *args, **kwargs):

    test_load_spectrum(plot=plot, verbose=verbose, warnings=warnings, *args, **kwargs)

    test_load_lines_pops(plot=plot, verbose=verbose, warnings=warnings, *args, **kwargs)

    return True


if __name__ == "__main__":

    printm("Testing database.py: ", _run_testcases(plot=True))



""""

Summary
-----------------------

Kurucz database parser

-----------------------

Defines :func:`~radis.io.fetch_kurucz` based on :class:`~radis.api.kuruczapi.KuruczDatabaseManager`

"""

from os.path import abspath, expanduser, join

import radis
from radis.api.kuruczapi import KuruczDatabaseManager
from radis.misc.config import getDatabankEntries


def fetch_kurucz(
    molecule,
    local_databases=None,
    databank_name="Kurucz-{molecule}",
    isotope=None,
    load_wavenum_min=None,
    load_wavenum_max=None,
    columns=None,
    cache=True,
    verbose=True,
    clean_cache_files=True,
    return_local_path=False,
    engine="default",
    output="pandas",
    parallel=True,
):
    """
    See e.g. :func:`~radis.io.hitemp.fetch_hitemp` for an explanation of the parameters largely applicable to `fetch_kurucz`

    .. note::

        if a registered entry already exists and `radis.config["ALLOW_OVERWRITE"]` is `True`:

        - if any situation arises where the databank needs to be re-downloaded, the possible urls are attempted in their usual order of preference, as if the databank hadn't been registered, rather than directly re-downloading from the same url that was previously registered, in case e.g. a new linelist has been uploaded since the databank was previously registered

        - If no partition function file is registered, e.g because one wasn't available server-side when the databank was last registered, an attempt is still made again to download it, to account for e.g. the case where one has since been uploaded
    """

    # largely based on :py:func:`~radis.io.fetch_geisa`
    if r"{molecule}" in databank_name:
        databank_name = databank_name.format(**{"molecule": molecule})

    if local_databases is None:

        local_databases = join(radis.config["DEFAULT_DOWNLOAD_PATH"], "kurucz")
    local_databases = abspath(expanduser(local_databases))

    ldb = KuruczDatabaseManager(
        databank_name,
        molecule=molecule,
        local_databases=local_databases,
        verbose=verbose,
        parallel=parallel,
        engine=engine,
    )

    local_files, urlnames = [], []
    pf_path = []
    if ldb.is_registered():
        entries = getDatabankEntries(ldb.name)
        local_files, urlnames = entries["path"], entries["download_url"]
        if "parfunc" in entries:
            pf_path = [entries["parfunc"]]

    get_pf_files = True
    if ldb.is_registered() and not radis.config["ALLOW_OVERWRITE"]:
        error = False
        if cache == "regen" or not local_files:
            error = True
        files_to_check = local_files + pf_path
        if ldb.get_missing_files(files_to_check):
            error = True
        if not pf_path:
            get_pf_files = False  # assume partition function file is either unavailable or undesired for current species
        if error:
            raise Exception(
                'Changes are required to the local database, and hence updating the registered entry, but "ALLOW_OVERWRITE" is False. Set `radis.config["ALLOW_OVERWRITE"]=True` to allow the changes to be made and config file to be automatically updated accordingly.'
            )

    # Delete files if needed:

    if cache == "regen":
        ldb.remove_local_files(local_files + pf_path)
    ldb.check_deprecated_files(
        ldb.get_existing_files(local_files),
        auto_remove=True if cache != "force" else False,
    )

    get_main_files = True

    if len(local_files) > 1 or len(urlnames) > 1:
        raise Exception("only 1 database file is expected")

    if local_files and not ldb.get_missing_files(local_files):
        get_main_files = False
        ldb.actual_file = local_files[0]  # for ldb.load below
    if pf_path and not ldb.get_missing_files(pf_path):
        get_pf_files = False
        ldb.pf_path = pf_path[0]

    # Download files
    if get_main_files:
        main_files, main_urls = ldb.get_possible_files()
        for i in range(len(main_urls)):
            url = main_urls[i]
            file = main_files[i]
            print(f"Attempting to download {url}")
            try:
                ldb.download_and_parse([url], [file], 1)
            except OSError:
                if i == len(main_urls) - 1:  # all possible urls exhausted
                    print(f"Error downloading {url}.")
                    print(f"No source found for {ldb.molecule}")
                    raise
                else:
                    print(f"Error downloading {url}")
                    continue
            else:
                print(f"Successfully downloaded {url}")
                ldb.actual_file = file
                ldb.actual_url = url
                # local_files = [file]
                break  # no need to search any further

    if get_pf_files:
        pf_path, pf_url = ldb.get_pf_path()
        ldb.pf_path = pf_path
        try:
            ldb.download_and_parse([pf_url], [pf_path], 1)
        except OSError:
            print("a partition function file specific to this species was not found")
            get_pf_files = False
            ldb.pf_path = None

    # Register
    if get_main_files or get_pf_files or not ldb.is_registered():
        ldb.register(get_main_files, get_pf_files)

    if (get_main_files or get_pf_files) and clean_cache_files:
        ldb.clean_download_files()

    if isotope and type(isotope) == int:
        isotope = str(isotope)

    # Load and return
    df = ldb.load(
        [ldb.actual_file],
        columns=columns,
        within=[("iso", isotope)] if isotope is not None else [],
        # for relevant files, get only the right range :
        lower_bound=[("wav", load_wavenum_min)] if load_wavenum_min is not None else [],
        upper_bound=[("wav", load_wavenum_max)] if load_wavenum_max is not None else [],
        output=output,
    )

    # based on ExoMol:
    if output in ["pandas", "vaex"]:  # no attribtes in "Jax" or "Vaex" mode
        attrs = {}
        attrs["molecule"] = molecule

        if output == "vaex":
            df.attrs = {}
            df.attrs = attrs
        elif output == "pandas":
            for k, v in attrs.items():
                df.attrs[k] = v

    return (df, local_files, ldb.pf_path) if return_local_path else df



# -*- coding: utf-8 -*-
"""
Created on Sun May 22 17:35:05 2022

@author: erwan
"""

from os.path import abspath, exists, expanduser, join

from radis.api.hdf5 import update_pytables_to_vaex
from radis.api.hitempapi import HITEMPDatabaseManager


def fetch_hitemp(
    molecule,
    local_databases=None,
    databank_name="HITEMP-{molecule}",
    isotope=None,
    load_wavenum_min=None,
    load_wavenum_max=None,
    columns=None,
    cache=True,
    verbose=True,
    chunksize=100000,
    clean_cache_files=True,
    return_local_path=False,
    engine="default",
    output="pandas",
    parallel=True,
):
    """Stream HITEMP file from HITRAN website. Unzip and build a HDF5 file directly.

    Returns a Pandas DataFrame containing all lines.

    Parameters
    ----------
    molecule: `"H2O", "CO2", "N2O", "CO", "CH4", "NO", "NO2", "OH"`
        HITEMP molecule. See https://hitran.org/hitemp/
    local_databases: str
        where to create the RADIS HDF5 files. Default ``"~/.radisdb/hitemp"``.
        Can be changed in ``radis.config["DEFAULT_DOWNLOAD_PATH"]`` or in ~/radis.json config file
    databank_name: str
        name of the databank in RADIS :ref:`Configuration file <label_lbl_config_file>`
        Default ``"HITEMP-{molecule}"``
    isotope: str, int or None
        load only certain isotopes : ``'2'``, ``'1,2'``, etc. If ``None``, loads
        everything. Default ``None``.
    load_wavenum_min, load_wavenum_max: float (cm-1)
        load only specific wavenumbers.
    columns: list of str
        list of columns to load. If ``None``, returns all columns in the file.

    Other Parameters
    ----------------
    cache: ``True``, ``False``, ``'regen'`` or ``'force'``
        if ``True``, use existing HDF5 file. If ``False`` or ``'regen'``, rebuild it.
        If ``'force'``, raise an error if cache file cannot be used (useful for debugging).
        Default ``True``.
    verbose: bool
    chunksize: int
        number of lines to process at a same time. Higher is usually faster
        but can create Memory problems and keep the user uninformed of the progress.
    clean_cache_files: bool
        if ``True`` clean downloaded cache files after HDF5 are created.
    return_local_path: bool
        if ``True``, also returns the path of the local database file.
    engine: 'pytables', 'vaex', 'default'
        which HDF5 library to use to parse local files. If 'default' use the value from ~/radis.json
    output: 'pandas', 'vaex', 'jax'
        format of the output DataFrame. If ``'jax'``, returns a dictionary of
        jax arrays. If ``'vaex'``, output is a :py:class:`vaex.dataframe.DataFrameLocal`

        .. note::
            Vaex DataFrames are memory-mapped. They do not take any space in RAM
            and are extremely useful to deal with the largest databases.

    parallel: bool
        if ``True``, uses joblib.parallel to load database with multiple processes

    Returns
    -------
    df: pd.DataFrame
        Line list
        A HDF5 file is also created in ``local_databases`` and referenced
        in the :ref:`RADIS config file <label_lbl_config_file>` with name
        ``databank_name``
    local_path: str
        path of local database file if ``return_local_path``

    Examples
    --------
    ::

        from radis import fetch_hitemp
        df = fetch_hitemp("CO")
        print(df.columns)
        >>> Index(['id', 'iso', 'wav', 'int', 'A', 'airbrd', 'selbrd', 'El', 'Tdpair',
            'Pshft', 'ierr', 'iref', 'lmix', 'gp', 'gpp', 'Fu', 'branch', 'jl',
            'syml', 'Fl', 'vu', 'vl'],
            dtype='object')

    .. minigallery:: radis.fetch_hitemp

    Notes
    -----
    if using ``load_only_wavenum_above/below`` or ``isotope``, the whole
    database is anyway downloaded and uncompressed to ``local_databases``
    fast access .HDF5 files (which will take a long time on first call). Only
    the expected wavenumber range & isotopes are returned. The .HFD5 parsing uses
    :py:func:`~radis.api.hdf5.hdf2df`

    See Also
    --------
    :py:func:`~radis.io.hitran.fetch_hitran`, :py:func:`~radis.io.exomol.fetch_exomol`
    :py:func:`~radis.io.geisa.fetch_geisa`, :py:func:`~radis.io.kurucz.fetch_kurucz`
    :py:func:`~radis.io.hdf5.hdf2df`
    :py:meth:`~radis.lbl.loader.DatabankLoader.fetch_databank`

    """

    if r"{molecule}" in databank_name:
        databank_name = databank_name.format(**{"molecule": molecule})

    if local_databases is None:
        import radis

        local_databases = join(radis.config["DEFAULT_DOWNLOAD_PATH"], "hitemp")
    local_databases = abspath(expanduser(local_databases))

    ldb = HITEMPDatabaseManager(
        databank_name,
        molecule=molecule,
        local_databases=local_databases,
        verbose=verbose,
        chunksize=chunksize,
        parallel=parallel,
        engine=engine,
    )

    # Get list of all expected local files for this database:
    local_files, urlnames = ldb.get_filenames()

    # Delete files if needed:
    relevant_files = ldb.keep_only_relevant(
        local_files, load_wavenum_min, load_wavenum_max, verbose=(verbose > 1)
    )
    if cache == "regen":
        ldb.remove_local_files(relevant_files)
    ldb.check_deprecated_files(
        ldb.get_existing_files(relevant_files),
        auto_remove=True if cache != "force" else False,
    )

    # Get missing files
    download_files = ldb.get_missing_files(local_files)
    download_files = ldb.keep_only_relevant(
        download_files, load_wavenum_min, load_wavenum_max, verbose=(verbose > 1)
    )
    # do not re-download files if they exist in another format :
    if engine in ["vaex", "auto", "default"]:
        # ... convert files if asked:
        from radis import config

        if config["AUTO_UPDATE_DATABASE"]:
            converted = []
            for f in download_files:
                if exists(f.replace(".hdf5", ".h5")):
                    update_pytables_to_vaex(f.replace(".hdf5", ".h5"))
                    converted.append(f)
            download_files = [f for f in download_files if f not in converted]
        # do not re-download remaining files that exist. Let user decide what to do.
        # (download & re-parsing is a long solution!)
        download_files = [
            f for f in download_files if not exists(f.replace(".hdf5", ".h5"))
        ]

    # Download files
    if len(download_files) > 0:
        if urlnames is None:
            urlnames = ldb.fetch_urlnames()
        filesmap = dict(zip(local_files, urlnames))
        download_urls = [filesmap[k] for k in download_files]
        ldb.download_and_parse(download_urls, download_files)

        # Done: add final checks
        if molecule not in ["CO2", "H2O"]:
            # check on the created file that all lines are there
            # ... (for CO2, H2O, database is split in many files so it's not possible)
            nrows = ldb.get_nrows(local_files[0])
            # assert nrows == Nlines
            _, Ntotal_lines_expected, _, _ = ldb.fetch_url_Nlines_wmin_wmax()
            if nrows != Ntotal_lines_expected:
                raise AssertionError(
                    f"Number of lines in local database ({nrows:,}) "
                    + "differ from the expected number of lines for "
                    + f"HITEMP {molecule}: {Ntotal_lines_expected}"
                )

    # Register
    if not ldb.is_registered():
        ldb.register()

    if len(download_files) > 0 and clean_cache_files:
        ldb.clean_download_files()

    # Load and return
    files_loaded = ldb.keep_only_relevant(
        local_files, load_wavenum_min, load_wavenum_max, verbose=verbose
    )

    if isotope and type(isotope) == int:
        isotope = str(isotope)

    df = ldb.load(
        files_loaded,  # filter other files,
        columns=columns,
        within=[("iso", isotope)] if isotope is not None else [],
        # for relevant files, get only the right range :
        lower_bound=[("wav", load_wavenum_min)] if load_wavenum_min is not None else [],
        upper_bound=[("wav", load_wavenum_max)] if load_wavenum_max is not None else [],
        output=output,
    )

    return (df, files_loaded) if return_local_path else df


if __name__ == "__main__":

    import pytest

    print("Testing factory:", pytest.main(["../test/io/test_hitemp.py"]))



# -*- coding: utf-8 -*-
"""
Test query functions

-------------------------------------------------------------------------------

"""

from os.path import exists, join

import pytest

from radis.misc.utils import NotInstalled, not_installed_vaex_args
from radis.misc.warning import DeprecatedFileWarning

try:
    import vaex
except ImportError:
    vaex = NotInstalled(*not_installed_vaex_args)

pytestmark = pytest.mark.random_order(disabled=True)

# ignored by pytest with argument -m "not needs_connection"
@pytest.mark.needs_connection
def test_fetch_astroquery(verbose=True, *args, **kwargs):
    """Test astroquery"""
    from radis.io.query import fetch_astroquery

    df = fetch_astroquery("CO2", 1, 2200, 2400, verbose=verbose, cache=False)

    assert df.iloc[0].id == 2
    assert df.iloc[0].iso == 1

    return


# ignored by pytest with argument -m "not needs_connection"
@pytest.mark.needs_connection
def test_fetch_astroquery_empty(verbose=True, *args, **kwargs):
    """Test astroquery: get a spectral range where there are no lines"""
    from radis.io.query import fetch_astroquery

    df = fetch_astroquery(
        2, 1, 25000, 50000, verbose=verbose, cache=False
    )  # 200-400 nm

    assert len(df) == 0

    return


# ignored by pytest with argument -m "not needs_connection"
@pytest.mark.needs_connection
def test_fetch_astroquery_cache(verbose=True, *args, **kwargs):
    """Test astroquery cache file generation

    - Cache file is generated
    - Cache file is loaded.
    - Different metadata raises an error
    """
    from astroquery.hitran import Hitran

    from radis.io.query import CACHE_FILE_NAME, fetch_astroquery

    df = fetch_astroquery(
        "CO2",
        1,
        2200,
        2400,
        verbose=verbose,
        cache="regen",  # Delete existing cache file if needed
    )

    # assert cache file was created
    assert exists(
        join(
            Hitran.cache_location,
            CACHE_FILE_NAME.format(
                **{"molecule": "CO2", "isotope": 1, "wmin": 2200, "wmax": 2400}
            ),
        )
    )

    # Try to load with different metadata: expect a DeprecatedFileError
    with pytest.raises(DeprecatedFileWarning):
        df2 = fetch_astroquery(
            "CO2",
            1,
            2200,
            2400,
            verbose=verbose,
            cache="force",  # force load of cache file
            expected_metadata={"not_in_metadata": True},
        )
    # Try to load with correct metadata. Expect to work and return the dataframe
    df2 = fetch_astroquery(
        "CO2",
        1,
        2200,
        2400,
        verbose=verbose,
        cache="force",  # force load of cache file
    )
    assert (df == df2).all().all()

    return


# ignored by pytest with argument -m "not needs_connection"
@pytest.mark.needs_connection
def test_fetch_hitran_CO_pytables(*args, **kwargs):

    from radis.io.hitran import fetch_hitran
    from radis.test.utils import getTestFile

    df = fetch_hitran(
        "CO",
        local_databases=join(getTestFile("."), "hitran"),
        databank_name="HITRAN-CO-TEST-ENGINE-PYTABLES",
        engine="pytables",
    )

    assert len(df) == 5381
    assert df.wav.min() == 3.40191
    assert df.wav.max() == 14477.377142


# TODO : clean database on each new pytest run ?


# ignored by pytest with argument -m "not needs_connection"
@pytest.mark.needs_connection
@pytest.mark.skipif(isinstance(vaex, NotInstalled), reason="Vaex not available")
def test_fetch_hitran_CO_vaex(*args, **kwargs):

    from radis.io.hitran import fetch_hitran
    from radis.test.utils import getTestFile

    df = fetch_hitran(
        "CO",
        local_databases=join(getTestFile("."), "hitran"),
        databank_name="HITRAN-CO-TEST-ENGINE-VAEX",
        engine="vaex",
    )

    assert len(df) == 5381
    assert df.wav.min() == 3.40191
    assert df.wav.max() == 14477.377142


# ignored by pytest with argument -m "not needs_connection"
@pytest.mark.needs_connection
def test_fetch_hitran(*args, **kwargs):

    from radis.io.hitran import fetch_hitran

    df = fetch_hitran("CO")

    assert len(df) == 5381
    assert df.wav.min() == 3.40191
    assert df.wav.max() == 14477.377142


@pytest.mark.needs_connection
def test_calc_hitran_spectrum(verbose=True, plot=False, *args, **kwargs):
    """
    Test full range & partial loading of CO spectrum
    """

    from radis import test_spectrum

    # The main thing here is to make sure the two function calls work :
    s = test_spectrum(databank=("hitran", "full"), name="full range", verbose=verbose)
    s2 = test_spectrum(
        databank=("hitran", "range"), name="partial range", verbose=verbose
    )

    if plot:
        import matplotlib.pyplot as plt

        from radis import plot_diff

        # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

        plot_diff(s, s2, method="ratio")

    # assert s == s2
    # Note: tiny differences (0.56%) probably due to how data is stored on disk
    assert s.compare_with(s2, spectra_only="abscoeff", rtol=0.007, plot=plot)

    return


@pytest.mark.needs_connection
@pytest.mark.skipif(isinstance(vaex, NotInstalled), reason="Vaex not available")
def test_pytable_vs_vaex(verbose=True):
    import numpy as np
    from astropy import units as u

    from radis import SpectrumFactory

    conditions = {
        "wmin": 2116 / u.cm,
        "wmax": 2123 / u.cm,
        "molecule": "CO",
        "isotope": "2",
        "pressure": 1.01325,  # bar
        "mole_fraction": 0.1,
        "path_length": 1,  # cm
        "broadening_method": "fft",
        "verbose": verbose,
    }
    Tgas = 1000
    sf_exo_pytables = SpectrumFactory(**conditions)
    sf_exo_pytables.fetch_databank(
        "exomol", memory_mapping_engine="pytables", db_use_cached="regen"
    )
    s_exo_pytables = sf_exo_pytables.eq_spectrum(Tgas=Tgas * u.K, path_length=1 * u.cm)

    sf_exo_vaex = SpectrumFactory(**conditions)
    sf_exo_vaex.fetch_databank(
        "exomol", memory_mapping_engine="vaex", db_use_cached="regen"
    )
    s_exo_vaex = sf_exo_vaex.eq_spectrum(Tgas=Tgas * u.K, path_length=1 * u.cm)

    assert np.isclose(
        s_exo_vaex.get_integral("abscoeff"),
        s_exo_pytables.get_integral("abscoeff"),
    )  # June 2024, should be 0.0026227144035907727


def _run_testcases(verbose=True, *args, **kwargs):

    test_fetch_astroquery(verbose=verbose, *args, **kwargs)
    test_fetch_astroquery_empty(verbose=verbose, *args, **kwargs)
    test_fetch_astroquery_cache(verbose=verbose, *args, **kwargs)
    test_fetch_hitran_CO_pytables(*args, **kwargs)
    test_fetch_hitran_CO_vaex(*args, **kwargs)
    test_fetch_hitran(*args, **kwargs)
    test_calc_hitran_spectrum(*args, **kwargs)
    test_pytable_vs_vaex(verbose=verbose)

    return True


if __name__ == "__main__":
    print("test_query.py: ", _run_testcases(verbose=True))



# -*- coding: utf-8 -*-
"""
Created on Fri Apr 14 21:10:31 2017

@author: erwan

Validation case
------------

Compare line broadening calculated in RADIS vs calculated in HAPI, using the
same database (HITRAN 2016)

"""

import shutil
from os.path import dirname, exists, join

import pytest

from radis import Spectrum, SpectrumFactory
from radis.db.classes import get_molecule_identifier
from radis.misc.printer import printm
from radis.phys.convert import nm2cm
from radis.test.utils import setup_test_line_databases


# @pytest.mark.fast not fast due to connection
@pytest.mark.needs_connection  # ignored by pytest with argument -m "not needs_http_connection"
def test_line_broadening(rtol=1e-3, verbose=True, plot=False, *args, **kwargs):
    r"""
    Plot absorption coefficient (cm-1) of CO at high temperature (2000 K) with
    RADIS, and compare with calculations from HAPI using the HITRAN database

    Notes
    -----

    In this example no data is needed. Everything is downloaded from the HITRAN
    database directly using either the HAPI ``fetch`` function, or the RADIS
    :meth:`~neq.spec.factory.fetch_databank` method.

    """

    from hapi import (
        absorptionCoefficient_Voigt,
        db_begin,
        fetch,
        tableList,
        transmittanceSpectrum,
    )

    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/radis.json if not there

    # Conditions
    molecule = "CO2"
    mol_id = get_molecule_identifier(molecule)
    iso = 1
    T = 1500
    p = 0.1
    L = 0.1
    #    M = 0.001           # mole fraction  (dont know where to put that )
    dnu = 0.0001
    wmin = nm2cm(4372.69 + 0.2)  # cm-1
    wmax = nm2cm(4372.69 - 0.2)  # cm-1
    truncation = 0.25  # cm-1
    neighbour_lines = 0.5  # cm-1

    # %% HITRAN calculation
    # -----------

    # Generate HAPI database locally

    HAPIdb = join(dirname(__file__), __file__.replace(".py", "_HAPIdata"))

    def calc_hapi():
        """Calc spectrum under HAPI"""

        clean_after_run = not exists(HAPIdb) and False

        try:
            db_begin(HAPIdb)
            if not molecule in tableList():  # only if data not downloaded already
                fetch(
                    molecule,
                    mol_id,
                    iso,
                    wmin - neighbour_lines,
                    wmax + neighbour_lines,
                )
                # HAPI doesnt correct for side effects

            # Calculate with HAPI
            nu, coef = absorptionCoefficient_Voigt(
                SourceTables="CO2",
                Environment={
                    "T": T,
                    "p": p / 1.01325,
                },  # K  # atm
                WavenumberStep=dnu,
                HITRAN_units=False,
                GammaL="gamma_self",
            )
            nu, trans = transmittanceSpectrum(
                nu,
                coef,
                Environment={
                    "l": L,
                },
            )  # cm
            s_hapi = Spectrum.from_array(
                nu,
                trans,
                "transmittance_noslit",
                "cm-1",
                "1",
                conditions={"Tgas": T},
                name="HAPI",
            )

        except:
            raise

        finally:
            if clean_after_run:
                shutil.rmtree(HAPIdb)
        return s_hapi

    s_hapi = calc_hapi()

    def calc_radis():

        # %% Calculate with RADIS
        # ----------
        pl = SpectrumFactory(
            wavenum_min=wmin,
            wavenum_max=wmax,
            mole_fraction=1,
            path_length=L,
            wstep=dnu,
            molecule=molecule,
            pressure=p,
            truncation=truncation,
            cutoff=1e-23,
            isotope=iso,
        )
        pl.warnings["MissingSelfBroadeningWarning"] = "ignore"
        pl.warnings["HighTemperatureWarning"] = "ignore"
        pl.fetch_databank(
            source="hitran",
            db_use_cached=True,
        )

        s = pl.eq_spectrum(Tgas=T)  # , Ttrans=300)
        s.name = "RADIS"

        if plot:
            pl.plot_broadening()

        return s

    s = calc_radis()

    # %% Compare
    # also shrinks HAPI range to the valid one
    s_hapi.resample(s.get_wavenumber(), unit="cm-1", energy_threshold=0.1)

    save = False  # just used in the article
    if plot or save:
        from radis import plot_diff

        #        title = '{0} bar, {1} K, {2} cm'.format(p, T, L)  if save else None
        fig, [ax0, ax1] = plot_diff(
            s, s_hapi, var="transmittance_noslit", method="ratio", show=plot
        )

        ax0.annotate(
            r"[P64](00$^\mathregular{0}$0)$\rightarrow $(00$^\mathregular{0}$1)",
            (2286.945, 0.76),
            (2286.94, 0.8),
            arrowprops=dict(arrowstyle="->", facecolor="black"),
        )
        ax0.annotate(
            r"[P53](01$^\mathregular{1}$0)$\rightarrow $(01$^\mathregular{1}$1)",
            (2286.9, 0.78),
            (2286.9, 0.82),
            arrowprops=dict(arrowstyle="->", facecolor="black"),
            horizontalalignment="right",
        )
        ax1.set_ylim(0.95, 1.05)

        if save:
            fig.savefig("out/test_RADIS_vs_HAPI_line_broadening.pdf")

    # Compare integrals
    diff = abs(
        s.get_integral("transmittance_noslit")
        / s_hapi.get_integral("transmittance_noslit")
        - 1
    )
    b = diff < rtol

    if verbose:
        printm(
            "Integral difference ({0:.2f}%) < {1:.2f}%: {2}".format(
                diff * 100, rtol * 100, b
            )
        )

    return b


if __name__ == "__main__":
    printm("test RADIS_vs_HAPI_line_broadening: ", test_line_broadening(plot=True))



# -*- coding: utf-8 -*-
"""Parsers for various databases

-------------------------------------------------------------------------------

"""

from .exomol import fetch_exomol
from .geisa import fetch_geisa
from .hitemp import fetch_hitemp
from .hitran import fetch_hitran
from .query import fetch_astroquery

__all__ = [
    "fetch_hitemp",
    "fetch_hitran",
    "fetch_exomol",
    "fetch_astroquery",
    "fetch_geisa",
]



import os.path
from ctypes import (
    POINTER,
    byref,
    c_char,
    c_int,
    c_longlong,
    c_short,
    c_void_p,
    cast,
    memmove,
    sizeof,
)
from os import name as os_name

if os_name == "nt":
    from ctypes import windll as dllobj
else:
    from ctypes import cdll as dllobj

from time import perf_counter

import numpy as np
from scipy.fft import irfft, rfft

from radis.gpu.structs import blockDim_t, gridDim_t
from radis.misc.utils import getProjectRoot

CUFFT_R2C = 0x2A
CUFFT_C2R = 0x2C


# This number does not have a meaningful interpretation in a CPU,
# so we just default to a typical number for GPU.
_max_threads_per_block = 1024


def getClasses():
    return CuContext, CuModule, CuArray, CuFFT, CuTimer


class CuContext:
    def __init__(self, device, context):
        # private:
        self._device = device
        self._context = context

    @staticmethod
    def setVerbosity(level):
        # Verbose output not implemented for GPU emulation
        pass

    @staticmethod
    def Open(device_id=0, flags=0, verbose=None):
        _device = c_void_p(456)
        _context = c_void_p(123)

        return CuContext(_device, _context)

    @staticmethod
    def getDeviceList():
        import platform

        return [platform.processor()]

    def printDeviceCapabilities(self):
        print("> GPU emulated by CPU")
        print("  Total amount of global memory:   0 bytes")

    def getMaxThreadsPerBlock(self):
        return _max_threads_per_block

    def setCurrent(self):
        pass

    def synchronize(self):
        pass

    def syncStream(self):
        pass

    def destroy(self):
        pass


class CuModule:
    def __init__(self, context, module_name):
        # public:
        lib_ext = ".dll" if os_name == "nt" else ".so"
        self.module_name = os.path.splitext(module_name)[0] + lib_ext
        self.context = context
        self.mode = "CPU"

        # private:
        # radis_path = dirname(dirname(__file__))
        radis_path = getProjectRoot()

        self._module = dllobj.LoadLibrary(
            os.path.join(radis_path, "gpu", "cuda", "build", self.module_name)
        )
        self._func_dict = {}
        self._global_dict = {}

    def __getattr__(self, attr):
        try:
            self._func_dict[attr]
            return self._func_dict[attr]

        except (KeyError):
            _function = getattr(self._module, attr)
            self._func_dict[attr] = CuFunction(_function)
            self._func_dict[attr].module = self
            return self._func_dict[attr]

    def _getGlobal(self, name, ctype=None):
        try:
            return self._global_dict[name]

        except (KeyError):
            if ctype is not None:
                _type = ctype
                _var = _type.in_dll(self._module, name)
                _size = sizeof(ctype)
                self._global_dict[name] = _var, _size, _type
                return self._global_dict[name]

            else:
                print("Constant type must be passed first time it is called!")
                return

    def getMode(self):
        return self.mode

    def setConstant(self, name, c_val):
        _var, _size, _type = self._getGlobal(name, type(c_val))
        memmove(byref(_var), byref(c_val), sizeof(c_val))

    def getConstant(self, name, ctype=None):
        _var, _size, _type = self._getGlobal(name, ctype=None)
        return _var


class CuFunction:
    def __init__(self, _function):

        # public:
        self.module = None
        self.args = []
        self.blocks = (1, 1, 1)
        self.threads = (1, 1, 1)
        self.sync = False

        # private:
        self._function = _function

    def setGrid(self, blocks=(1, 1, 1), threads=(1, 1, 1)):
        self.blocks = blocks
        self.threads = threads

    def setArgs(self, *vargs):
        self.args = vargs

    def __call__(self, *vargs, blocks=None, threads=None, sync=None):

        self.args = self.args if not len(vargs) else vargs
        self.blocks = self.blocks if blocks is None else blocks
        self.threads = self.threads if threads is None else threads
        self.sync = self.sync if sync is None else sync

        self.module.setConstant("blockDim", blockDim_t(*self.threads))
        self.module.setConstant("gridDim", gridDim_t(*self.blocks))

        c_args = [arr._ptr for arr in self.args]
        self._function(*c_args)


class CuArray:
    def __init__(self, shape, dtype=np.float32, init="empty", grow_only=False):
        self._ptr = c_void_p()
        self.grow_only = grow_only  # Note that this doesn't actually do anything!
        self.resize(shape, dtype, init)

    def resize(self, shape=None, dtype=None, init="empty"):
        shape_tuple = shape if isinstance(shape, tuple) else (shape,)
        self.shape = self.shape if shape is None else shape_tuple
        self.dtype = self.dtype if dtype is None else np.dtype(dtype)
        self.size = int(np.prod(shape))
        self.itemsize = self.dtype.itemsize
        self.nbytes = self.itemsize * self.size

        if init not in ("zeros", "empty"):
            return

        elif init == "empty":
            self._arr = np.empty(self.shape, dtype=self.dtype)

        elif init == "zeros":
            self._arr = np.zeros(self.shape, dtype=self.dtype)

        self._ptr = c_void_p(self._arr.ctypes.data)

    @staticmethod
    def fromArray(arr):

        obj = CuArray(arr.shape, arr.dtype, init=None)
        obj.setArray(arr)

        return obj

    def zeroFill(self):
        self.getArray()[:] = 0

    def setArray(self, arr):

        params_changed = arr.shape != self.shape or arr.dtype != self.dtype
        uninitialized_memory = self._ptr.value is None
        if params_changed or uninitialized_memory:
            self.resize(arr.shape, arr.dtype, None)
        self._ptr = c_void_p(arr.ctypes.data)
        self._arr = arr

    def getArray(self):
        my_ctype = {1: c_char, 2: c_short, 4: c_int, 8: c_longlong}[self.itemsize]
        my_cptr = cast(self._ptr.value, POINTER(my_ctype))
        arr = np.ctypeslib.as_array(my_cptr, self.shape).view(self.dtype)
        return arr


class CuFFT:
    def __init__(self, arr_in, arr_out, workarea=None, direction="fwd"):

        # public:
        self.arr_in = arr_in
        self.arr_out = arr_out
        self.workarea = CuArray(0, dtype=np.byte) if workarea is None else workarea

        # private:
        self._direction = direction
        self._arr = arr_in if direction == "fwd" else arr_out
        self._fft_type = CUFFT_R2C if direction == "fwd" else CUFFT_C2R
        self._plans = {}

    def __call__(self):

        np_arr_in = self.arr_in.getArray()
        np_arr_out = self.arr_out.getArray()

        if self._direction == "fwd":
            np_arr_out[:] = rfft(np_arr_in, axis=0, workers=-1)
        else:
            n = np_arr_out.shape[0]
            np_arr_out[:] = irfft(np_arr_in, n=n, axis=0, workers=-1) * n

    def destroy(self):
        pass


class CuTimer:
    def __init__(self, stream=0):
        self._stream = c_int(stream)
        self._start = perf_counter()
        self._stop = self._start
        self.times = {}

    def reset(self):
        self._start = perf_counter()

    def lap(self, name=None):
        if name is None:
            name = "event{:d}".format(len(self.times.keys()))
        self.times[name] = self()

    def __call__(self):
        self._stop = perf_counter()
        elapsed_time = 1e3 * (self._stop - self._start)
        return elapsed_time

    def getTimes(self):
        return self.times

    def getDiffs(self):
        vals = [*self.times.values()]
        diffs = [vals[0]] + [vals[i] - vals[i - 1] for i in range(1, len(vals))]
        return dict(zip(self.times.keys(), diffs))
